/* ACCES-mPCIe-DIO-24S.c - ACCES I/O Products DIO driver test program */

/* Copyright (c) 2018, ACCES I/O Products, Inc.
 * 10623 Roselle St, San Diego CA, 92121-1506 USA
 * +1-858-550-9559 | +1-800-326-1649 | sales@accessio.com
 * 
 * The right to copy, distribute, modify, or otherwise make use of this software 
 * may be pursuant to the terms of an applicable ACCES I/O Products, Inc. 
 * (ACCES) license agreement, or in the absence of such agreement, is limited to 
 * use with ACCES hardware products.
 */

/*
modification history
--------------------
21jun18,acc  Initial release.
*/

/*
DESCRIPTION

Test program for ACCES I/O Products DIO cards driver.

*/

#include <vxWorks.h>
#include <private/iosLibP.h>
#include <hwif/buslib/pciDefines.h>
#include <stdio.h>
#include <string.h>
#include <usrLib.h>                     /* for devs() */
#include <hwif/vxBus.h>
#include <hwif/vxbus/vxbBusType.h>
#include <subsys/int/vxbIntLib.h>
#include <hwif/buslib/vxbPciLib.h>

#include "vxbAccesDio.h"


IMPORT SL_LIST vxbPciCtrlList;
IMPORT BOOL vxbPciCtrlListInited;

/* No header file for these functions? */
IMPORT void sysMsDelay (int);   /* delays int msecs */
IMPORT void vxbPciCtrlShow();
IMPORT void vxbPciDeviceShow(VXB_DEV_ID pciBusCtrlId, UINT8 busNo);

/* #define ACCES_DEBUG */
#ifdef ACCES_DEBUG
/* Functions in the driver for use in test. */
IMPORT void   accesDioDebugLevelSet (int);
#endif /* ACCES_DEBUG */

/* Forwards */
VXB_DEV_ID  vxbPciBusCtrlIdGet(void);

/* "Local" variables made global for debug. */
UINT32      accesDioPorts;
VXB_DEV_ID  pciBusCtrlId;      /* The PCI bus. */

#define NO_ARG_CALL_TEST(FNAME)                                     \
do { if (FNAME () == OK)                                            \
        printf (#FNAME "() OK.\n");                                 \
    else                                                            \
        {                                                           \
        printf (#FNAME "() failed with errno = %d.\n", errno);      \
        goto L_err;                                                 \
        }                                                           \
    }                                                               \
while (0)

int accesDioTestISRcount;

/******************************************************************************
*
* accesDioTestISR - simple ISR for interrupt test--just increment a count
*/

STATUS accesDioTestISR (void)
    {
    accesDioTestISRcount++;
    return OK;
    }

/******************************************************************************
*
* accesDioIntTest - test DIO card Change-of-State interrupts
* 
* Do all steps to enable interrupts, wait 30 seconds for the user to short a
* pin, and report # interrupts received. A quick paper-clip short will generate
* a few dozen interrupts. See code for details.
*/

STATUS accesDioIntTest (void)
    {
    /***** TEST: Interrupts. */ 

    /* Establish our ISR callback function and enable interrupts for 30 seconds. */

    NO_ARG_CALL_TEST(accesDioOpen);
    
    if (accesDioIntCallbackInstall (accesDioTestISR) == ERROR)
        { printf ("accesDioIntTest:%d: failed with errno = %d.\n", __LINE__, errno); goto L_err; }

    /* Remember! Ports must be set for input for interrupts to work. */
    if (accesDioConfigure(PORT_DIR_INPUT, PORT_DIR_INPUT, PORT_DIR_INPUT, PORT_DIR_INPUT) == OK)
        printf ("accesDioConfigure() OK\n");
    else
        { printf ("accesDioTest:%d: failed with errno = %d.\n", __LINE__, errno); goto L_err; }
    
    if (accesDioIntEnable () == ERROR)
        { printf ("accesDioIntTest:%d: failed with errno = %d.\n", __LINE__, errno); goto L_err; }

    printf ("Set ISR call back function and enable interrupts for 30 seconds. Short any DIO\n"
            "pin (A: 30-37), B (22-29), C (3-10) to ground (1) to cause an interrupt.\n"
            "Our callback function, accesDioTestISR(), increments a local counter which\n"
            "is reported at the end of the test.\n");
    
    sysMsDelay (30000);
    
    NO_ARG_CALL_TEST(accesDioClose);
    
    printf ("Interrupt test done. accesDioTestISRcount = %d\n", accesDioTestISRcount);

    return OK;
    
    /* In Workbench, place a breakpoint here to break on errors. */
L_err:
    printf ("ERROR: At L_err label. errno if relevant = %d.\n", errno);
    return ERROR;
    }

/******************************************************************************
*
* accesDioTest - test DIO driver non-interrupt functions
* 
* Main test for DIO card. Flash LEDs in various ways. See code for details.
*/

STATUS accesDioTest (void)
    {
    UINT8   controlWord;
    int     cycle;
    UINT32  pinCnt;
    UINT32  ports;
    
#ifdef ACCES_DEBUG
    accesDioDebugLevelSet (0);
#endif /* ACCES_DEBUG */
    
    /* Open the card and configure all ports for output. */
    
    NO_ARG_CALL_TEST(accesDioOpen);
    
    if (accesDioConfigure(PORT_DIR_OUTPUT, PORT_DIR_OUTPUT, PORT_DIR_OUTPUT, PORT_DIR_OUTPUT) == OK)
        printf ("accesDioConfigure() OK\n");
    else
        { printf ("accesDioTest:%d: failed with errno = %d.\n", __LINE__, errno); goto L_err; }
    
    accesDioReadControl (&controlWord);
    printf ("Control Word = 0x%04x\n", controlWord );
    
    /***** TEST: Toggle all 24 LEDs left to right and back twice. */

    for (cycle = 1; cycle <= 2; cycle++)
        {
        ports = 1;
        for (pinCnt = 1; pinCnt <= 24; pinCnt++)
            {
            accesDioWritePorts (ports);
            ports = ports << 1;
            sysMsDelay (100);
            }
        
        for (pinCnt = 24; pinCnt >= 1; pinCnt--)
            {
            accesDioWritePorts (ports);
            ports = ports >> 1;
            sysMsDelay (100);
            }
        }
    
    /***** TEST: Set bits on, wait 3 seconds, set off, wait 3 seconds; do twice. */

    printf ("Set all bits off, wait 3 seconds, set all on, wait 3 seconds; do twice.\n");

    for (cycle = 1; cycle <= 2; cycle++)
        {
        printf ("  on ...");
        if (accesDioWritePorts (0xffffff) == ERROR)
            { printf ("accesDioTest:%d: failed with errno = %d.\n", __LINE__, errno); goto L_err; }
        sysMsDelay (3000);
    
        printf ("  off ...");
        if (accesDioWritePorts (0) == ERROR)
            { printf ("accesDioTest:%d: failed with errno = %d.\n", __LINE__, errno); goto L_err; }
        sysMsDelay (3000);
        }
    printf ("\n");
    
    /***** TEST: Toggle all bits from 1 to 0 and back at 10 Hz for 10 seconds. */ 

    printf ("DIO blink test start: blinking LEDs on DIO at 10 Hz for 10 sec\n");
    
    for (cycle=1; cycle <= 50; cycle++)
        {
        if (accesDioWritePorts (0xffffff) == ERROR)
            { printf ("accesDioTest:%d: failed with errno = %d.\n", __LINE__, errno); goto L_err; }
        sysMsDelay (100);
        if (accesDioWritePorts (0) == ERROR)
            { printf ("accesDioTest:%d: failed with errno = %d.\n", __LINE__, errno); goto L_err; }
        sysMsDelay (100);
        }
    
    printf ("DIO blink test done.\n");

    NO_ARG_CALL_TEST(accesDioClose);
    
    printf ("accesDioTest Done.\n");
    return OK;
    
    /* In Workbench, place a breakpoint here to break on errors. */
L_err:
    printf ("ERROR: At L_err label. errno if relevant = %d.\n", errno);
    return ERROR;
    }

/******************************************************************************
*
* accesDioPciBusesShow - show devices on each PCI bus and pause before each bus
*/

STATUS accesDioPciBusesShow()
    {
    VXB_DEV_ID      pciBusCtrlId;
    UINT8           busNo;
    char            response[10];        /* For nnn\n + safety */
    
    if ((pciBusCtrlId = vxbPciBusCtrlIdGet()) == NULL) return ERROR;
    
    printf ("At each $ prompt, press Enter for next bus (starting at 0), Ctrl+d to terminate,"
          " 0 <= n <= 255 for bus n.\n");
    
    for (busNo = 0; busNo <= 255; busNo++)
        {
        printf ("$ ");
        if (gets (response) == NULL) return OK;
        
        if (strlen(response) > 0)
            busNo = atoi (response);
        
        vxbPciDeviceShow (pciBusCtrlId, busNo);
        }
    
    return OK;  /* Avoid gcc warning. Won't get here. busNo is UINT8 so 255+1 = 0. */ 
    
    }

/******************************************************************************
*
* vxbPciBusCtrlIdGet - return PCI bus controller pointer for bus tests
*/

VXB_DEV_ID vxbPciBusCtrlIdGet(void)
    {
    SL_NODE *       pNode;
    VXB_PCI_CTRL *  pPciCtrl = NULL;

    if (!vxbPciCtrlListInited) return NULL;

    vxbPciCtrlShow();

    pNode = SLL_FIRST (&vxbPciCtrlList);
    pPciCtrl = (VXB_PCI_CTRL *)pNode;
    pciBusCtrlId = pPciCtrl->pDev;
    
    printf("vxbPciBusCtrlIdGet: %p.", pciBusCtrlId);
    return pciBusCtrlId;
    }
