/* ACCES-mPCIe_COM_4SM.c - Test program for ACCES I/O Products COM cards */

/* Copyright (c) 2018, ACCES I/O Products, Inc.
 * 10623 Roselle St, San Diego CA, 92121-1506 USA
 * +1-858-550-9559 | +1-800-326-1649 | sales@accessio.com
 * 
 * The right to copy, distribute, modify, or otherwise make use of this software 
 * may be pursuant to the terms of an applicable ACCES I/O Products, Inc. 
 * (ACCES) license agreement, or in the absence of such agreement, is limited to 
 * use with ACCES hardware products.
 */
    
/*
modification history
--------------------
19may18,acc  first release
*/

/*
Contents
========
Description
    Requirements
    Hardware and ports
Installation and Execution
    Files
    Install
    Build and run
ACCES-mPCIE_COM_4SM Program
    Terminology
    Operation
    Commands, current tyCo
    Error handling
    Naming convention
Lower-Level Testing
    
Description
===========
The ACCES I/O Products mPCIe-COM-4SM is a four-port serial I/O card with
support for RS-232 up to 921.6K bps and RS-422 and RS-485 up to 3 Mbps.
 
This file is a test program for the card and associated VxWorks 7 serial 
driver. The program has several commands to exercise the card--see "Commands"
below.

Requirements
------------
The four ports of the card are referred to generically as Port A - Port D
(see "Hardware and Ports" below).

  *****************************************************************************
- Ports A-D must be discovered by the VxWorks PCI bus initialization function
  and assigned as devices /tyCo/1 - /tyCo/4. A "base" tyCo could be added.
  *****************************************************************************

- /tyCo/1 must be connected to a terminal or terminal emulator for test command
  input and output. Thus, the mPCIe-COM-4SM and driver must be working, at  
  at least for /tyCo/1 (see "Lower-Level Testing" below if it is not).

- Protocol attributes must be: 115200 baud, no parity, 8 data bits, and 1 stop 
  bit. (The 115200 baud rate requirement comes from the default value in the 
  NS16550_CFG structs in vxbPciNs16550Sio.c and may be changed there).

- This test program and the drivers were developed with the VxWorks 7 
  $WIND_BASE/pkgs/os/board/ia/itl_64_vx7-1.1.1.0 BSP. It should work with the
  $WIND_BASE/pkgs/os/board/ia/itl_generic-1.0.5.1 BSP.

NOTE: the driver is not dependent on interface type and any of RS-232, RS-422,
or RS-485 may be used on any of the four ports (the terminal or emulator on   
/tyCo/1 must be connected to a compatible port).

Hardware and ports
------------------
The four ports on the mPCIe-COM-4SM are identified generically as Port A
through Port D. When facing the port cable end of the card with the label on
the UART chip up, the ports and their corresponding device name and channel
number are arranged top/bottom left/right as follows:

    Upper left                          Upper right
    --------------------------          --------------------------
    Port A : /tyCo/3 : channel          Port B : /tyCo/4 : channel 
    
    Lower left                          Lower right
    --------------------------          --------------------------
    Port D : /tyCo/1 : channel          Port C : /tyCo/2 : channel 


Installation and Execution
==========================

Files
-----
NOTE: The reader is assumed to be familiar with the VxWorks directory structure
and VSB-VIP build processes.

Three files comprise this test program and driver:

- ACCES-mPCIE_COM_4SM.c:
    This test program.

- vxbPciNs16550Sio.c:        
    Replacement for standard VxWorks "National Semiconductor 16550 Serial PCI
    Driver" file. The changes extend the port configuration data and the 
    pciNs16550Probe() and pciNs16550Attach() functions for the mPCIe-COM-4SM.

- vxbNs16550Sio.c:            
    Standard VxWorks "National Semiconductor 16550 Serial Driver" with one-line
    modification required to support more than one channel (port) on a device.

Install
-------
- Install the mPCIe-COM-4SM. See the ACCES I/O Products "Multi-Port Serial for
  Mini PCI Express Hardware Manual" for details.
- Place this file ACCES-mPCIE_COM_4SM.c in the VIP directory.
- Replace file 
  $WIND_BASE/pkgs/os/drv/vxbus-1.0.5.0/drv-1.2.7.1/src/sio/vxbPciNs16550Sio.c
  with the delivered file. 
- Replace file 
  $WIND_BASE/pkgs/os/drv/vxbus-1.0.5.0/drv-1.2.7.1/src/sio/vxbNs16550Sio.c with
  the delivered file.

Build and run
-------------
- Create a VSB and VIP. Placing this file in your VIP directory causes it to
  compile and link.
- Transfer the resulting vxWorks binary image to your target.
- At the VxWorks kernel shell prompt enter
      -> accesComTest
  and operate the test program per "Commands" below.

accesComTest Program
====================
Terminology
-----------
- display       Write to the PC Console.
- Console       Short form for "PC Console"; use "Console" in general in case
                it changes from the PC Console. All command input and output 
                is on the Console in this program.   
- PC Console    Device /pcConsole/0: the "PC Console" in VxWorks kernel 
                configuration terms. 
- read          Read from a tyCo.
- request       Prompt for and get user input from the Console.
- terminal      The device at the remote end of a cable connected to an
                mPCIe-COM-4SM port, typically a terminal emulator on a host.
- write     Write to a tyCo.
- tyCo      One of /tyCo/1 ... /tyCo/4 or all per context.

Operation
---------
The main routine is accesComTest() and it can be run only after VxWorks has 
booted, typically by entering accesComTest at the VxWorks kernel shell. 

Before starting the program, enter the devs command to the shell. Assuming no
other serial cards in the system, the four mPCIe-COM-4SM ports will show as
/tyCo/1 through /tyCo/4 (see above for association with cables; /tyCo/0 is 
setup by VxWorks by default but is not connected if using the DuraCOR 311).

accesComTest() when started:
- Displays a greeting and a help message on the Console and pauses. 
- It then enters a command loop, requesting commands from the console. Commands 
  are all single letter, lower-case; some have an argument as shown.

Commands, current tyCo
----------------------
Command input and output is ALWAYS from/to the console, which is never a tyCo.
All output written by commands (e.g., test strings) is to the "current tyCo".

The "current tyCo" is set with the c command.

Writing to the current tyCo may make it easier to capture the output (e.g.,
on a terminal with a log capability, vs. a VGA PC Console). 

Common initial testing order: t, e, l. 

Commands (the implementing function is shown after each command):

b rate
    Change baud rate for the current tyCo to the given rate. If the current
    tyCo is connected to a terminal, change the rate at the terminal 
    immediately after entering the requested baud rate and pressing Enter.
    Then do command t or e, for example, to try the new rate. baudSet().

c [n]
    Set the "current tyCo" [module global curTyCoNameAcces] to the given n 
    (1-4). Default: /tyCo/1. If n is absent, report the current tyCo.

e   Echo test: read a character from the current tyCo and immediately echo it
    until that character is end-of-line (CR or LF), then re-write the whole 
    line to the current tyCo. A line may be up to 1000 characters (checked).
    echoTest().

h   Display command help.
   
p   Write the NS16550VXB_CHAN struct for each channel to the current tyCo. 
    pChanDisplay().

s   Write a 41 character test string to all four tyCos. simpleTest().

z   Exit the command loop and return from this function.

Error handling
--------------
In this test code, we want to break immediately on any error related to the 
/tyCo/n UARTs on the ACCES mPCIe-COM-4SM card. Thus, the return value of every  
operation to a tyCo is tested immediately after the call. If the return value 
indicates failure, our ERRacces() macro,
  - Calls printf() to write an error message to Console. Each message begins 
    with "*** FAILURE in __func__:__LINE__: error: " to help distinguish it 
    from any other output and locate it.
  - Calls our global function errorAcces() here. error() does a longjmp() to 
    return from this program (e.g., back to the VxWorks kernel shell). You can 
    break on error() to stop the program before the longjmp() and examine the 
    stack and global variables by entering VxWorks kernel shell command:
        b error
    or by setting a breakpoint at error() in the Workbench Debugger.
    
NOTE: It is assumed that input/output to the Console work. The program begins
with I/O to the console so a failure there can be debugged early.

Naming convention
-----------------
Every global variable and function in this file begins or ends with "acces" or
"Acces" to avoid clash with any VxWorks or user application names.  

Lower-Level Testing
===================
This program requires that the mPCIe-COM-4SM and drivers are working correctly
because it relies on Port A (/tyCo/1) for command and test input and output.

If there is doubt about either the mPCIe-COM-4SM or drivers, once the VxWorks
kernel shell is running, the following commands, executed at the shell prompt,
may be useful. These commands may also be of interest even if this test program
is able to communicate with /tyCo/1.  

- devs      Display a list of devices. accesMpcieCom4sm () starts with this. 

- fd=open("/tyCo/n",2)
            Open /tyCo/n (n=1-4) for read/write (second argument) and assign
            the resulting file descriptor to new symbol fd.
            
            Example:
                fd=open("/tyCo/3",2)
            
- write(fd,"abc",3)
            Write string "abc" to fd (must be open). The third argument must
            be the number of characters to write.
            
- mybuf=malloc(100)
            Allocate a 100 byte buffer and assign it to mybuf. The shell will
            report the address.

- read(fd, address reported for mybuf, n)
            Read n characters from fd into the buffer at the given address.
            The address must be a number (e.g., 0xffff800000XXXX); otherwise
            mybuff will be evaluated first and the read done to the address
            that is whatever is then in the first 8 bytes of mybuf.
            
- ifShow    Show network interfaces.

- d [adr][, num][, width]
            Dump memory at address adr for num units of width. adr must be
            numeric; width must be 1, 2, 4, or 8.
            
            Example: dump 3 4-byte words at xxxxxxxxxxxxxx
            d xxxxxxxxxxx,3,4
*/

#include <vxWorks.h>
#if defined(PRJ_BUILD)
#include "prjParams.h"
#endif /* defined PRJ_BUILD */

#include <fioLib.h>             /* For fdprintf(). */
#include <ioLib.h>
#include <private/iosLibP.h>    /* For close(), open(), read(), write(). */
#include <setjmp.h>             /* See "Error handling" above. */
#include <stdio.h>              /* For kprintf(), snprintf(), sprintf(). */
#include <string.h>             /* For strlen(). */
#include <tyLib.h>              /* For TY_DEV, used below in TYCO_DEV. */
#include <usrLib.h>             /* Kernel shell functions; for devs(). */
#include <hwif/vxbus/vxbLib.h>

/*
 * ERR macro: Display a message on the Console by calling fdprintf() and then 
 * call error() (see it). As with printf(), the first argument should be a
 * format string and any remaining arguements the values required by it.
 * 
 * The message will always start with:
 * 
 *  *** FAILURE in <function name>:<line number>: "
 *
 * Approach follows that explained at 
 * https://stackoverflow.com/questions/1644868/c-define-macro-for-debug-printing
 * Modified to output to the Console (rather than stdErr) because we want any 
 * output to the Console to interleave line-by-line with the error output. Also 
 * removed __FILE__ (not needed for this test program), and not conditional a  
 * debug flag.
 * 
 * NOTE: Use ERRacces() when the program cannot continue; use CONPRINT for ordinary
 * output or error messages to the Console when the program can continue (e.g.,
 * to report an invalid user input for a command).
 */
#define ERRacces(fmt, ...) \
            do { fdprintf(fdConsoleAcces, "*** FAILURE in %s():%d: " fmt, \
                          __func__, __LINE__, ##__VA_ARGS__); errorAcces(); } while (0)

/*
 * CONPRINT macro: Convenience macro to call fdprintf() for Console output. 
 * "fdprintf(fdConsoleAcces, ...)". Arguments are as for printf().
 */
#define CONPRINT(fmt, ...) fdprintf (fdConsoleAcces, fmt, ##__VA_ARGS__)

/*
 * TYCO_DEV is required to access the pointer to the NS16550VXB_CHAN "driver 
 * controller struct" created in vxbPciNs16550Sio.c: pciNs16550Attach() and
 * functions it calls. There and in vxbPciNs16550Sio.c this pointer is named
 * pChanAcces; in higher-level functions in tyLib.h, it is SIO_CHAN pSioChan.
 * TYCO_DEV is declared in ttyDrv.c but not made public. The definition here
 * is copied from there except pSioChan is given type (void *) to avoid having
 * to include its type's definition from sioLib.h (which is overloaded and 
 * misleading by TYCO_DEV).
 */
typedef struct tyCoDev
    {
    TY_DEV  tyDev;          /* First member of TY_DEV is DEV_HDR struct. */
    void *  pSioChan;       /* Pointer to  NS16550VXB_CHAN for the channel. */
    } TYCO_DEV;

/*
 * Most "local" variables used only in this file are made global so that each
 * can be examined in the VxWorks target shell C interpreter or Workbench 
 * Debugger.
 */
char        chrAcces;            /* For line read char-by-char in echoTest() */
char        cmdNameAcces[10];    /* Command name from cmdLineAcces. */
char        cmdLineAcces[100];   /* Command line requested from user. */
char        curTyCoNameAcces[20];/* Name of tyCo selected by user with s command;
                                    8 holds "/tyCo/n" + NUL; extra to avoid
                                    errors and 20 is used in drivers. */
char        errMsgAcces[1000];   /* Large error message buffer. */
int         fdConsoleAcces;      /* Console file FD. */
int         fdTyCo1Acces;        /* File descriptor for /tyCo/1, etc. */
int         fdTyCo2Acces;
int         fdTyCo3Acces;
int         fdTyCo4Acces;
int         iFdAcces;            /* Index to FD loop. */
jmp_buf     jmpBufEnvAcces;      /* For use by longjmp() in error(). */
char        lineAcces[1000];     /* Line read from the /tyCo/n test channel. */          
ssize_t     iLineAcces;          /* Index to echoTestAcces() test lineAcces. */
void *      pChanAcces;          /* To call vxbPciNs16550SioChannelShow(). */
int *       pCurTyCoFdAcces;     /* Pointer to FD for current tyCo. */  
VXB_DEV_ID  pDevAcces;           /* To call vxbPciNs16550SioChannelShow(). */
DEV_HDR *   pDevHdrAcces;        /* To call vxbPciNs16550SioChannelShow(). */
FD_ENTRY *  pFdEntryAcces;       /* To call vxbPciNs16550SioChannelShow(). */
TYCO_DEV *  pTyCoDevAcces;       /* To get pChanAcces; see TYCO_DEV above. */
STATUS      statusAcces;         /* Return from some system calls. */

/* Forward declarations. */

void errorAcces (void);
void baudSetAcces (void);        
void currentTyCoSetAcces (void);        
void echoTestAcces (void);       
void helpShowAcces (void);       
void loopbackTestAcces (void);   
void pChanDisplayAcces (void);   
void simpleTestAcces (void);        
     

/* 
 * Show() function included in vxbPciNs16550Sio.c to output the contents of an
 * NS16550VXB_CHAN struct to file descriptor fd. The 2nd argument is for
 * identification and is display only. The 3rd argument is the pChan pointer.
 */
void    vxbPciNs16550SioChannelShow (int fd, int channelN, void *);

/******************************************************************************
*
* accesComTest - test the ACCES I/O Products mPCIe-COM-4SM card
* 
* This is the main function for this program. See "accesComTest Program" above.
*/ 

STATUS accesComTest (void)
    {
    fdConsoleAcces = open("/pcConsole/0", O_RDWR, 0);
    if (ioctl (fdConsoleAcces, 
               FIOSETOPTIONS, OPT_LINE | OPT_ECHO | OPT_CRMOD ) == ERROR)
        ERRacces ("setting ioctl OPT_LINE option for Console.");
    
    setjmp (jmpBufEnvAcces);         /* Prepare for error() calls. */
	
    /* Display greeting on Console, show devices. */
	
    CONPRINT ("\n"
              "******************************************************************************\n"
              "ACCES mPCIe-COM-4SM Test Program. Copyright 2018 ACCES I/O Products, Inc.\n"
              "******************************************************************************\n"
              "See header in ACCES-mPCIE_COM_4SM.c for description. Current devices are shown\n"
              "below (/tyCo/1 - /tyCo/4 should appear for the ports of the mPCIe-COM-4SM).\n"
              "Enter 'h' for help. Enter z at the command prompt to exit this program and\n"
              "return to the VxWorks shell. The default \"current tyCo\" for tests is /tyCo/1;\n"
              "change it with the c command.\n\n");
    
    devs ();
    
    CONPRINT ("\nPress Enter to continue: ");
    (void)getchar();
    
    /* Open /tyCo/ devices. */ 
    
    if ((fdTyCo1Acces = open ("/tyCo/1", O_RDWR, 0)) == ERROR)
        ERRacces ("open(\"/tyCo/1\") returned %i.\n", fdTyCo1Acces);
    
    if ((fdTyCo2Acces = open ("/tyCo/2", O_RDWR, 0)) == ERROR)
        ERRacces ("open(\"/tyCo/2\") returned %i.\n", fdTyCo2Acces);

    if ((fdTyCo3Acces = open ("/tyCo/3", O_RDWR, 0)) == ERROR)
        ERRacces ("open(\"/tyCo/3\") returned %i.\n", fdTyCo3Acces);

    if ((fdTyCo4Acces = open ("/tyCo/4", O_RDWR, 0)) == ERROR)
        ERRacces ("open(\"/tyCo/4\") returned %i.\n", fdTyCo4Acces);

    /* Set up default "current tyCo". */
    pCurTyCoFdAcces = &fdTyCo1Acces;
    strcpy (curTyCoNameAcces, "/tyCo/1");

    /* Command loop on Console. */
    
    helpShowAcces ();
    
    for (;;)
        {
        CONPRINT("> ");
        read (fdConsoleAcces, cmdLineAcces, NELEMENTS(cmdLineAcces));

        switch (cmdLineAcces[0])
            {
            case 'b': baudSetAcces ();       break;
            case 'c': currentTyCoSetAcces ();break;
            case 'e': echoTestAcces ();      break;
            case 'h': helpShowAcces ();      break;
            case 'p': pChanDisplayAcces ();  break;
            case 's': simpleTestAcces ();    break;
            case 'z': 
                {
                CONPRINT ("\n\nGood bye.\n");

                close (fdTyCo1Acces);
                close (fdTyCo2Acces);
                close (fdTyCo3Acces);
                close (fdTyCo4Acces);
                
                close(fdConsoleAcces);
                
                return OK;
                }
            case '\x0d':    /* CR */
            case '\x0a':    /* LF */
                break;
                
            default:
                CONPRINT ("Invalid command letter '%c'. Enter 'h' for help.\n", 
                          cmdLineAcces[0]);
            }
        }
    }

/******************************************************************************
*
* baudSetAcces - change baud rate to requested tyCo to value from command line.
* 
* Command: b <rate>
* 
* Permitted baud rates are: 110, 300, 600, 1200, 2400, 4800, 9600, 14400, 19200, 
* 38400, 57600, and 115200.
* 
* NOTES:
* - VxWorks ioctl() has a check for a max baud rate of 115200: 
*   pkgs/os/drv/vxbus-1.0.5.0/drv-1.2.7.1/src/sio/vxbNs16550Sio.c has
*       #define NS16550_MAX_RATE 115200
*   This could be changed. Since we are delivering this file, consider changing
*   it.
*   
* - Other baud rates within the above range are possible and are supported by
*   Tera Term (a popular terminal emulator) by typing the rate rather than 
*   selecting it from the drop-down, but limiting to the Tera Term set is
*   sufficient.
*/

void baudSetAcces (void)
    {
    int         status;
    UINT32      baudRate;
    
    if (sscanf(&cmdLineAcces[1], "%u", &baudRate) != 1)
        {
        CONPRINT ("Invalid command line \"%s\"; must be \"b <baud rate>\""
                  " where <baud rate> is digits only.\n", 
                  cmdLineAcces);
        return;
        }
    
    /*
     * switch is an easy way to check the input value. NOTES: other baud rates
     * are possible  
     */
    switch (baudRate)
        {
        case 110: case 300: case 600: case 1200: case 2400: case 4800:
        case 9600: case 14400: case 19200: case 38400: case 57600:
        case 115200: break;
        
        default: 
            CONPRINT ("Invalid baud rate %u; accepted rates are\n"); 
            CONPRINT ("  110, 300, 600, 1200, 2400, 4800, 9600, 14400, 19200,"
                      " 38400, 57600, and 115200.\n");       
            return;
        }
    
    if ((status = ioctl (*pCurTyCoFdAcces, FIOBAUDRATE, baudRate)) == ERROR)
        {
        CONPRINT ("ioctl() to set baud rate returned error %i.\n", status);
        return;
        }
    
    fdprintf (fdConsoleAcces, "Baud rate for %s set to %u.\n", 
              curTyCoNameAcces, baudRate);
    }

/******************************************************************************
*
* currentTyCoSetAcces - set "current tyCo" (curTyCoNameAcces) from command line
*
* Command: c [<n>]
* 
* If <n> is not present, report the current tyCo.
*/

 void currentTyCoSetAcces (void)
     {
     UINT   n;
     int    sscanfResult;
     
     sscanfResult = sscanf(&cmdLineAcces[1], "%u", &n);
     
     if (sscanfResult == EOF)
         {
         CONPRINT ("Current tyCo = %s\n", curTyCoNameAcces);
         return;
         }
     
     else if (sscanfResult != 1)
         {
         CONPRINT ("Invalid command line \"%s\"; must be \"c n\" where n = 1-4.\n",
                   cmdLineAcces);
         return;
         }
     
     switch (n)
         {
         case 1: pCurTyCoFdAcces = &fdTyCo1Acces; break;
         case 2: pCurTyCoFdAcces = &fdTyCo2Acces; break;
         case 3: pCurTyCoFdAcces = &fdTyCo3Acces; break;
         case 4: pCurTyCoFdAcces = &fdTyCo4Acces; break;
         
         default:
             CONPRINT ("Invalid number following command c; must be 1-4.\n");
             return;
         }
    
     sprintf (curTyCoNameAcces, "/tyCo/%u", n);
     CONPRINT ("%s selected.\n", curTyCoNameAcces);
     }
 
/******************************************************************************
*
* echoTestAcces - read a character from a tyCo and write it until end-of-line (EOL)
*
* EOL is CR or LF; at EOL, re-write the whole line. A line may be up to 1000
* characters (checked). The tyCo used is set by request.
*/
    
void echoTestAcces (void)
    {
    const char  CTRL_Z = '\x1a';
    ssize_t     status;
    
    fdprintf(fdConsoleAcces, "Starting echo loop on %s. Enter Ctrl+Z there to "
                            "return here.\n", curTyCoNameAcces);
    
    if (fdprintf (*pCurTyCoFdAcces,
                   "\n\n          Hello. Welcome to ACCES I/O Products Echo!\n\n"    
                   "Enter lines terminated with CR or LF.\n"
                   "Characters are echoed as you type and the line is\n"
                   "re-output when terminated. Enter Ctrl+Z to terminate.\n")
                    < 0)
        ERRacces ("printf() of Hello message to %s failed with error %d.\n", 
                 curTyCoNameAcces); 

    
    FOREVER /* read/echo line loop */ 
        {
        if (fdprintf (*pCurTyCoFdAcces, "\n$ ") == ERROR)     /* "$ " prompt. */
            ERRacces ("fdprintf() of prompt '$ ' to %s failed.\n", curTyCoNameAcces); 
            
        iLineAcces = 0;
        FOREVER /* read/echo character loop */
            {
            if ((status = read (*pCurTyCoFdAcces, &chrAcces, 1)) == ERROR)
                ERRacces ("read() on %s failed with error %i.\n", 
                         curTyCoNameAcces, status); 
            
            if (chrAcces == CTRL_Z)                          /* End echo test? */
                {
                if ((status = write (*pCurTyCoFdAcces, "^Z\r\n", 4)) == ERROR)
                    ERRacces ("write(\"%s\", %c, 1) to echo ^Z failed with"
                        " error %i.\n", curTyCoNameAcces, status); 
                
                if (fdprintf (*pCurTyCoFdAcces, "Returning to command loop on Console.\n") == ERROR)
                    ERRacces ("fdprintf() of echoTestAcces() exit message to %s failed.\n", curTyCoNameAcces); 
                
                return;
                }
            
            if ((chrAcces == '\n') || (chrAcces == '\r'))     /* End of line? */ 
                {
                if (fdprintf (*pCurTyCoFdAcces, "\n= ") == ERROR)
                    ERRacces ("write() \"= \" to echo line to %s failed with error %i.\n", 
                         curTyCoNameAcces, status); 

                if ((status = write (*pCurTyCoFdAcces, lineAcces, iLineAcces)) == ERROR)
                    ERRacces ("write() to echo line to %s failed with error %i.\n", 
                             curTyCoNameAcces, status); 
                
                if ((status = write (*pCurTyCoFdAcces, "\n", 1)) == ERROR)
                    ERRacces ("write() to echo '\\n' at end of echo line to %s failed with error %i.\n", 
                         curTyCoNameAcces, status);
                break;
                }
            else    /* Ordinary character--add to line if room and echo. */
                {
                if (iLineAcces == NELEMENTS(lineAcces))
                    ERRacces ("at max line length (%i). Calling errorAcces ().",
                              NELEMENTS(lineAcces));
                
                lineAcces[iLineAcces++] = chrAcces;
    
                if ((status = write (*pCurTyCoFdAcces, &chrAcces, 1)) == ERROR)
                    ERRacces ("write(\"%s\", \"%c\", 1) to echo character failed"
                        " with error %i.\n", curTyCoNameAcces, chrAcces, status); 
                }
            } /* read/echo character loop */
        } /* read/echo line loop */ 
    }

/******************************************************************************
*
* errorAcces - Do-nothing function to provide a common location for a breakpoint
* 
* Do-nothing function to provide a global address to insert a breakpoint
* using the VxWorks target shell or Workbench Debugger. If no break, does
* a longjmp() to the environment set up by setjmp(jmpBufEnvAcces). 
*
* See "Error handling" above for details.
*/

void errorAcces (void)
    {
    setjmp(jmpBufEnvAcces);
    }

/******************************************************************************
*
* helpShowAcces - show help on the Console.
*/
    
void helpShowAcces (void)
    {
    CONPRINT ("\nACCES mPCIe-COM-4SM accesComTest Commands:\n\n");
    CONPRINT ("b rate\n"
              "    Change baud rate on current tyCo to the given rate.\n\n");
    CONPRINT ("c [n]\n");
    CONPRINT ("    Set the \"current tyCo\" to the given n (1-4). Default: /tyCo/1.\n");
    CONPRINT ("    If n is absent, display the current tyCo.\n\n");
    CONPRINT ("e   Echo test: read and echo a character from the current tyCo until\n");
    CONPRINT ("    end-of-line (CR or LF), then re-write the whole line. A line may be\n");
    CONPRINT ("    up to 1000 characters.\n\n");
    CONPRINT ("p   Write the NS16550VXB_CHAN struct for each channel on current tyCo.\n\n");
    CONPRINT ("s   Simple test: write a 41 character test string to all four tyCos..\n\n");
    CONPRINT ("z   Exit the command loop and return from this function.\n\n");
    }

 /******************************************************************************
 *
 * pChanDisplayAcces - display each NS16550VXB_CHAN struct on the Console
 */

 void pChanDisplayAcces (void)
     {
     /* 
      * Develop pFdEntryAcces, pDevHdrAcces, pTyCoDevAcces, and pChanAcces for 
      * clarity and debug. A DEV_HDR struct is the first member of TYCO_DEV
      * and the final member is our pChan pointer. 
      */

     for (iFdAcces = fdTyCo1Acces; iFdAcces <= fdTyCo4Acces; iFdAcces++)
         {
         pFdEntryAcces = iosFdMap (iFdAcces);
         pDevHdrAcces  = (DEV_HDR *) pFdEntryAcces->pDevHdr;
         pTyCoDevAcces = (TYCO_DEV *) pDevHdrAcces; 
         pChanAcces    = pTyCoDevAcces->pSioChan;

         vxbPciNs16550SioChannelShow (*pCurTyCoFdAcces, iFdAcces - fdTyCo1Acces, 
                                      pChanAcces);
         }
     }


/******************************************************************************
*
* simpleTestAcces - write a 41 character test string to all four tyCos
* 
* This test helps identify which device (/tyCo/1, etc.) corresponds to which 
* connector on the mPCIE-COM-4SM card by writing a self-identifying string to
* all four ports. Connect a cable in turn from each port to the terminal to
* identify the device associated with that port.
*/
    
void simpleTestAcces (void)
    {
    if (fdprintf (fdTyCo1Acces, "\n Test1 Test1 Test1 Test1 Test1 Test1\n") == ERROR) 
        ERRacces ("simpleTest: Error on write to /tyCo/1.");
    else
        CONPRINT ("Sent test string to /tyCo/1");
    
    if (fdprintf (fdTyCo2Acces, "\n Test2 Test2 Test2 Test2 Test2 Test2\n") == ERROR) 
        ERRacces ("simpleTest: Error on write to /tyCo/2.");
    else
        CONPRINT (", /tyCo/2");

    if (fdprintf (fdTyCo3Acces, "\n Test3 Test3 Test3 Test3 Test3 Test3\n") == ERROR) 
        ERRacces ("simpleTest: Error on write to /tyCo/3.");
    else
        CONPRINT (", /tyCo/3");

    if (fdprintf (fdTyCo4Acces, "\n Test4 Test4 Test4 Test4 Test4 Test4\n") == ERROR) 
        ERRacces ("simpleTest: Error on write to /tyCo/4.");
    else
        CONPRINT (", /tyCo/4.\n");
    }
