/* multiboot.c -  multiboot information structure library */

/*
 * Copyright 2014-2015 (c) Wind River Systems, Inc.
 *
 * The right to copy, distribute, modify or otherwise make use
 * of this software may be licensed only pursuant to the terms
 * of an applicable Wind River license agreement.
 */

/*
modification history
--------------------
15sep15,j_l  Fix compile warning
19aug15,dot  V7PRO-2102 - fixed apigen errors and warnings.
30jan15,scm  V7PRO-1656 multibootShow initialization issues
04nov14,rbc  US44720 - Fix static analysis errors
13aug14,ten  US44071 - Merge multiboot changes from itl_64 BSP:
                       US38341, V7PRO-891
27mar14,j_l  US30676 - remove config.h
05mar14,jjk  WXV7-2052 - Support multiple markers for bootline pre-pending
31jan14,j_l  initial creation based on itl_64 version 01aug13
*/

/*
DESCRIPTION
An initial attempt to understand the contents of the multiboot
information structure passed to the operating system by
the multiboot loader.  mem and mmap fields might be useful
in satisfying sysPhysMemTop
*/

#include <vxWorks.h>
#include <pc.h>
#include <arch/i86/x86_64/private/adrSpaceArchLibP.h>

/* multiboot defined states */

#define MMAR_DESC_TYPE_AVAILABLE    1 /* available to OS */
#define MMAR_DESC_TYPE_RESERVED     2 /* not available */
#define MMAR_DESC_TYPE_ACPI_RECLAIM 3 /* usable by OS after reading ACPI */
#define MMAR_DESC_TYPE_ACPI_NVS     4 /* required to save between NVS */
                                      /* sessions */

#define PAGE_SIZE4K 4096

typedef struct mb_info
    {
    unsigned int  flags;
    unsigned int  mem_lower;
    unsigned int  mem_upper;
    unsigned int  boot_device;
    unsigned int  cmdline;
    unsigned int  mods_count;
    unsigned int  mods_addr;
    unsigned int  syms[4];
    unsigned int  mmap_length;
    unsigned int  mmap_addr;
    unsigned int  drives_length;
    unsigned int  drives_addr;
    unsigned int  config_table;
    unsigned int  boot_loader_name;
    unsigned int  apm_table;
    unsigned int  vbe_control_info;
    unsigned int  vbe_mode_info;
    unsigned int  vbe_mode;
    unsigned int  vbe_interface_seg;
    unsigned int  vbe_interface_off;
    unsigned int  vbe_interface_len;
    } MB_INFO;

typedef struct mmap_buf
    {
    unsigned int  size;
    unsigned int  base_addr_low;
    unsigned int  base_addr_high;
    unsigned int  length_low;
    unsigned int  length_high;
    unsigned int  type;
    } MMAP_BUF;

IMPORT char * sysBootLine;

const MB_INFO* pMbInfo = (MB_INFO *)(~(0UL));

void multibootInit ()
    {
    UINT32 *p = (UINT32 *)(MULTIBOOT_SCRATCH);

    if (*p != MULTIBOOT_BOOTLOADER_MAGIC)
        return;

    p = (UINT32 *)(MULTIBOOT_SCRATCH + 8);

    pMbInfo = (MB_INFO*)((ULONG)(*p));

    if (pMbInfo->flags & 4)
        {
        char * bootline;

        /*
         * Search for prepended string '*bootline:'. This will match
         * 'sysbootline:' as well as 'bootline:',
         *
         * Otherwise set to DEFAULT_BOOT_LINE...
         */

        bootline = strstr((char *)((ULONG)pMbInfo->cmdline), "bootline:");
        if (NULL != bootline)
            strncpy (sysBootLine,
                     (char *)(bootline + strlen("bootline:")), 255);
        else
            strncpy(sysBootLine,
                    DEFAULT_BOOT_LINE, strlen(DEFAULT_BOOT_LINE)+1);
        }
    }


/*******************************************************************************
*
* multibootMapDescGet - get a multiboot memory map descriptor
*
* This routine finds the descriptor that corresponds to the requested index,
* and copies it to the user provided storage.
*
*/

STATUS multibootMapDescGet
    (
    unsigned int index,
    MMAP_BUF * pMapBuf
    )
    {
    int entries;
    MMAP_BUF * pMap;

    /* verify table */

    if (pMbInfo == (MB_INFO *)(~(0UL)))
        return (ERROR);

    if ((pMbInfo->flags & 0x40) == 0)
        return (ERROR);

    /* verify parameters */

    entries = pMbInfo->mmap_length / sizeof(MMAP_BUF);

    if (index > entries)
        return (ERROR);

    pMap = (MMAP_BUF *)(uintptr_t)pMbInfo->mmap_addr;
    *pMapBuf = pMap[index];

    return (OK);
    }


/*******************************************************************************
*
* vxBiosE820MapGet64 - get the multiboot memory map
*
* The routine populates the 64 bit physical ram descriptor table using values
* from the memory map in the multiboot info structure. This routine will
* locate existing entries in sysPhysRamDesc and update the length value with
* the values reported in the corresponding multiboot memory map entry.  If
* an existing sysPhysRamDesc[] entry is not found, a new one will be generated
* using the next available dummy entry.
*
* All existing entries that do not correspond to system-reported memory regions
* will not be modified.  The sysPhysRamDesc table must be populated with enough
* dummy entries to satisfy each available memory region reported in the system
* memory descriptor table.
*
*/

STATUS multibootMapGet64
    (
    PHYS_MEM_DESC *sysPhysRamDesc,
    int sysPhysRamDescCount
    )
    {
    unsigned int i;
    unsigned int ram_entry = 0;
    unsigned int usable = 0;
    unsigned int dummy_entries = 0;
    unsigned int match_entries = 0;
    BOOL found = FALSE;
    unsigned int entries;
    MMAP_BUF * pMap;
    UINT64 addr;
    UINT64 length;

    if (pMbInfo == (MB_INFO *)(~(0UL)))
        return (ERROR);

    /* verify parameters */

    if ((sysPhysRamDesc == NULL) || (sysPhysRamDescCount <= 0))
        return (ERROR);

    /* verify table */

    if (pMbInfo == (MB_INFO *)(~(0UL)))
        return (ERROR);

    if ((pMbInfo->flags & 0x40) == 0)
        return (ERROR);

    /*
     * Count available dummy entries in sysPhysRamDesc.
     * A descriptor with a length value of 0 denotes a dummy entry.
     */

    for (i=0; i<sysPhysRamDescCount; i++,ram_entry++)
        {
        if (sysPhysRamDesc[ram_entry].len != 0)
            continue;
        dummy_entries++;
        }

    entries = pMbInfo->mmap_length / sizeof(MMAP_BUF);
    pMap = (MMAP_BUF *)(uintptr_t)pMbInfo->mmap_addr;

    /* Count usable entries, and matching entries in sysPhyMemDesc */

    for (i=0; i < entries; i++)
        {

        addr = pMap[i].base_addr_low | ((UINT64)pMap[i].base_addr_high << 32);
        length = pMap[i].length_low | ((UINT64)pMap[i].length_high << 32);

        /* ignore unusable entries and the first segment at address 0x0 */

        if ((pMap[i].type != MMAR_DESC_TYPE_AVAILABLE) || (addr == 0x0))
            continue;

        for (ram_entry=0;ram_entry<sysPhysRamDescCount;ram_entry++)
            {
            if (addr == sysPhysRamDesc[ram_entry].physicalAddr)
                {
                match_entries++;
                break;
                }
            }
        usable++;
        }

    /* check available space in sysPhysMemDesc table */

    if (usable > (match_entries+dummy_entries))
        {

        /* not enough space in the table */

        return (ERROR);
        }

    /* space requirements have been satisfied, now update the table */

    for (i=0; i < entries; i++)
        {
        found = FALSE;

        addr = pMap[i].base_addr_low | ((UINT64)pMap[i].base_addr_high << 32);
        length = pMap[i].length_low | ((UINT64)pMap[i].length_high << 32);

        /* ignore unusable entries and the first segment at address 0x0 */

        if ((pMap[i].type != MMAR_DESC_TYPE_AVAILABLE) || (addr == 0x0))
            continue;

        if ((addr & (PAGE_SIZE4K - 1)) != 0)
            {
            addr += PAGE_SIZE4K;
            length -= PAGE_SIZE4K;
            addr &= ~(PAGE_SIZE4K - 1);
            }

        if ((length & (PAGE_SIZE4K - 1)) != 0)
            {
            length &= ~(PAGE_SIZE4K - 1);
            }

        for (ram_entry=0;ram_entry<sysPhysRamDescCount;ram_entry++)
            {

            /* adjust any existing segment to actual size */

            if (addr == sysPhysRamDesc[ram_entry].physicalAddr)
                {
                sysPhysRamDesc[ram_entry].len = length;
                found = TRUE;
                break;
                }
            }

        if (!found)
            {

            /*
             * Now add unmapped segments -
             * find the next available dummy slot and
             * populate the RAM descriptor entry
             */

            for (ram_entry=0;ram_entry<sysPhysRamDescCount;ram_entry++)
                {
                if (sysPhysRamDesc[ram_entry].len == 0)
                    {
                    sysPhysRamDesc[ram_entry].physicalAddr = addr;
                    sysPhysRamDesc[ram_entry].len = length;
                    break;
                    }
                }
            }
        }

    return (OK);
    }

#define MULTIBOOT_DEBUG
#ifdef  MULTIBOOT_DEBUG
#include <stdio.h>
void multibootShow ()
    {
    unsigned int flags;

    UINT32 *p = (UINT32 *)((MULTIBOOT_SCRATCH) | (KERNEL_DIRECT_MAP_RGN_BASE));

    if (*p != MULTIBOOT_BOOTLOADER_MAGIC)
      return;

    p = (UINT32 *)((MULTIBOOT_SCRATCH + 8) | (KERNEL_DIRECT_MAP_RGN_BASE));

    pMbInfo = (MB_INFO*)(((ULONG)(*p)) | (KERNEL_DIRECT_MAP_RGN_BASE));

    flags = (unsigned int)pMbInfo->flags;
    (void) printf ("flags: 0x%x\n", flags);

    if (flags & 1)
        (void) printf ("mem_lower: 0x%x mem_upper: 0x%x\n",
                       (unsigned int)pMbInfo->mem_lower,
                       (unsigned int)pMbInfo->mem_upper);

    if (flags & 2)
        (void) printf ("boot_device: 0x%x\n",
                       (unsigned int) pMbInfo->boot_device);

    if (flags & 4)
        (void) printf ("cmdline: 0x%x\n", (unsigned int) pMbInfo->cmdline);

    if (flags & 8)
        {
        (void) printf ("mods_length: 0x%x\n", pMbInfo->mods_count);
        (void) printf ("mods_addr: 0x%x\n", pMbInfo->mods_addr);
        }

    if (flags & 0x40)
        {
        int i;
        char* mmap_area;
        (void) printf ("mmap_length: 0x%x\n", pMbInfo->mmap_length);
        (void) printf ("mmap_addr: 0x%x\n", pMbInfo->mmap_addr);
        (void) printf ("mmap_areas type   start address                 "
                       "region length\n");

        mmap_area = (char*) (((ULONG)pMbInfo->mmap_addr) |
                             (KERNEL_DIRECT_MAP_RGN_BASE));
        for (i = 0; i < pMbInfo->mmap_length;)
            {
            MMAP_BUF* mmap = (MMAP_BUF*)mmap_area;
            (void) printf("region      %d    0x%.8x:%.8x         0x%.8x:%.8x\n",
                   mmap->type,
                   mmap->base_addr_high,
                   mmap->base_addr_low,
                   mmap->length_high,
                   mmap->length_low);
            i += mmap->size + sizeof (unsigned int);
            mmap_area += (mmap->size + sizeof (unsigned int));
            }
        }

    if (flags & 0x80)
        {
        (void) printf ("drives_length: 0x%x\n", pMbInfo->drives_length);
        (void) printf ("drives_addr: 0x%x\n", pMbInfo->drives_addr);
        }
    }
#endif /* MULTIBOOT_DEBUG */
