/* romInit.s - ROM initialization module */

/*
 * Copyright (c) 2014, 2015-2016 Wind River Systems, Inc.
 *
 * The right to copy, distribute, modify or otherwise make use
 * of this software may be licensed only pursuant to the terms
 * of an applicable Wind River license agreement.
 */

/*
modification history
--------------------
15aug16,emj  fix apigen synopsis tags
19aug15,dot  V7PRO-2102 - fixed apigen errors and warnings.
13aug14,ten  US44071 - Merge multiboot changes from itl_64 BSP:
                       US38341, V7PRO-891
27mar14,j_l  US30676 - remove config.h
31jan14,jjk  DE5023 - Do not duplicate source in VSB.
31jan14,j_l  initial creation based on itl_64 version 01aug13
*/

/*
DESCRIPTION
This module contains the entry code for the VxWorks bootrom.

The routine sysToMonitor(2) jumps to the location XXX bytes
passed the beginning of romInit, to perform a "warm boot".

This code is intended to be generic accross i80x86 boards.
Hardware that requires special register setting or memory
mapping to be done immediately, may do so here.

The following is a diagram that illustrates the general sequence of
boot/warmboot processor states.

\cs
For Single Stage Boot:

                                            warm boot
                           +<------------------------------------+
                           |                                     ^
                           |                                     |
                           |
                           |             +--> ELF32 --> b32 --> v32
                           |             |
                           v             |
                                         |
vxld (16 bit) --> b16 --> b32 --> b64 -->+
                                         |
                                   ^     |
                                   |     +--> ELF64 ----------> v64
                                   |
                                   |                             |
                                   |                             v
                                   +<----------------------------+
                                            warm boot


For Multi Stage Boot:

                            warm boot (with  INCLUDE_FAST_REBOOT)
                           +<------------------------------------+
                           |                                     ^
                           |                                     |
                           |             +--> ELF32 --> b32 --> v32
                           v             |
vxld (16 bit) --> b16 --> b32 --> b64 -->+
                           ^             |
                           |             +--> ELF64 ----------> v64
                           |                                     |
                           |                                     v
                           +<---- v32 <--------------------------+
                             warm boot (with INCLUDE_FAST_REBOOT)

                   warm boot (with INCLUDE_MULTI_STAGE_WARM_REBOOT)
                   +<--------------------------------------------+
                   |                                             ^
                   |                                             |
                   |                     +--> ELF32 --> b32 --> v32
                   v                     |
vxld (16 bit) --> b16 --> b32 --> b64 -->+
                   ^                     |
                   |                     +--> ELF64 ----------> v64
                   |                                             |
                   |                                             v
                   +<---- b32 <-- v32 <--------------------------+
                   warm boot (with INCLUDE_MULTI_STAGE_WARM_REBOOT)


Key
===
b16  : bootrom/bootApp (CPU in 16 bit mode)
b32  : bootrom/bootApp (CPU in 32 bit mode)
b64  : bootrom/bootApp (CPU in 64 bit mode)
         loads vxWorks image at this stage
ELF  : bootrom/bootApp (CPU in 32 bit mode)
         vxWorks 32 bit ELF32 file has loaded
ELF32: bootrom/bootApp (CPU in 64 bit mode)
         determines vxWorks 32 bit ELF32 file has loaded
ELF64: bootrom/bootApp (CPU in 64 bit mode)
         determines vxWorks 64 bit ELF64 file has loaded
v32  : vxWorks 32 bit  (CPU in 32 bit mode)
v64  : vxWorks 64 bit  (CPU in 64 bit mode)
\ce

*/

#define _ASMLANGUAGE
#include <vxWorks.h>
#include <sysLib.h>
#include <asm.h>

#include <pc.h>

#include <regs.h>

#define SP_ARG1         4

	.data

	/* internals */

	.globl	romInit			/* start of system code */
	.globl	_romInit		/* start of system code */

	.globl	romTableStart
	.globl	_romTableStart
	.globl	romTableEnd
	.globl	_romTableEnd

	.globl	sdata			/* start of data */
	.globl	_sdata			/* start of data */

        /* externals */

        .globl  FUNC(romA20on)         /* turn on A20 */

 sdata:
_sdata:
	.asciz	"start of data"

	.text
	.balign 16

        .code32

/*******************************************************************************
*
* romInit - entry point for VxWorks in ROM
*
* SYNOPSIS
* \ss
* romInit (startType)
*     int startType;	/@ only used by 2nd entry point @/
* \se
*
*/

	/* cold start entry point in REAL MODE(16 bits) */

romInit:
_romInit:
	cli				/* LOCK INTERRUPT */

    /*
     * check if loaded by multiboot (in protected mode) or vxld (in real mode), 
     * then jump to the corresponding start for that mode.
     *
     * Note: it was decided that ROM_WARM_HIGH should always be at offset
     * 0x10 from the start of romInit(), and in order for that to happen,
     * the romInit() entry point code must fit in 16 bytes. Unfortunately,
     * because the cold() entry point is so far away, a non-relative
     * jump is required to do a "jmp cold" which, in conjunction with the
     * .code16/.addr16 prefix bytes, puts us a few bytes over the limit.
     * The original solution for this was just to move ROM_WARM_HIGH,
     * ROM_WARM_LOW and everything else down another 16 bytes. However
     * this is unacceptable as it makes this BSP inconsistent with all
     * the other IA BSPs.
     *
     * Instead, an extra trampoline stage for 16-bit cold start has been
     * added at offset 0x30, which is close enough to allow for a relative
     * jump instruction. This allows the code below to fit in exactly
     * 16 bytes which avoids the need to change ROM_WARM_HIGH. The
     * zeroing of the %ecx register could probably be eliminated as well,
     * since %ecx will be completely overwritten by %cr0 anyway, which
     * would save another couple of bytes. But for now this isn't needed.
     */

	movl	%cr0, %ecx
	and	$0x01, %cx
	jnz	cold32
	.byte 0x67, 0x66
	jmp	coldtramp		/* offset must be less than 128 */


	/* warm start entry point in PROTECTED MODE(32 bits) */

	.balign 16,0x90
romWarmHigh:				/* ROM_WARM_HIGH(0x10) is offset */
	cli				/* LOCK INTERRUPT */
	movl    SP_ARG1(%esp),%ebx	/* %ebx has the startType */
					/* If coming from 64 bit, edi has startType */
	jmp	warm

	/* warm start entry point in PROTECTED MODE(32 bits) */

	.balign 16,0x90
romWarmLow:				/* ROM_WARM_LOW(0x20) is offset */
	cli				/* LOCK INTERRUPT */
	movl	%edi, %eax		/* If coming from 64 bit */
					/*   save edi, the startType */
	cld				/* copy itself to ROM_TEXT_ADRS */
	movl	$ SYS_RAM_LOW_ADRS,%esi	/* get src addr (RAM_LOW_ADRS) */
	movl	$ ROM_TEXT_ADRS,%edi	/* get dst addr (ROM_TEXT_ADRS) */
	movl	$ ROM_SIZE,%ecx		/* get nBytes to copy */
	shrl	$2,%ecx			/* get nLongs to copy */
	rep				/* repeat next inst ECX time */
	movsl				/* copy nLongs from src to dst */
	movl    SP_ARG1(%esp),%ebx	/* %ebx has the startType */
	movl	%eax, %edi		/* If coming from 64 bit */
					/*   restore edi, the startType */
	jmp	warm			/* jump to warm */

	/* cold start entry point in REAL MODE (16 bits) */

	.balign 16, 0x90
coldtramp:
	.byte 0x67, 0x66
	jmp cold

	/* cold start entry point in PROTECTED MODE (32 bits) */

	.balign 16, 0x90
cold32:
#ifdef GRUB_MULTIBOOT
        mov     %eax,(MULTIBOOT_SCRATCH)     /* save magic */
        mov     %ebx,(MULTIBOOT_SCRATCH + 8) /* save multi-boot pointer */
#endif /* GRUB_MULTIBOOT */
	movl    SP_ARG1(%esp),%ebx	/* %ebx has the startType */
	jmp	warm

#ifdef GRUB_MULTIBOOT
        /* Multiboot Header allows GRUB to load kernel */
        .balign 4
multiboot_header:
        .long   MULTIBOOT_HEADER_MAGIC
        .long   MULTIBOOT_HEADER_FLAGS
        .long   -(MULTIBOOT_HEADER_MAGIC + MULTIBOOT_HEADER_FLAGS)
        .long   multiboot_header + ROM_TEXT_ADRS - romInit
        .long   ROM_TEXT_ADRS
        .long   wrs_kernel_data_end + ROM_TEXT_ADRS - romInit
        .long   wrs_kernel_bss_end + ROM_TEXT_ADRS - romInit
        .long   cold32
#endif /* GRUB_MULTIBOOT */

	/* copyright notice appears at beginning of ROM (in TEXT segment) */

	.ascii   "Copyright 2011 Wind River Systems, Inc."


	/* cold start code in REAL MODE(16 bits) */

	.balign 16,0x90
cold:

#ifdef GRUB_MULTIBOOT
        mov     %eax,(MULTIBOOT_SCRATCH)     /* save magic */
        mov     %ebx,(MULTIBOOT_SCRATCH + 8) /* save multi-boot pointer */
#endif /* GRUB_MULTIBOOT */

        /*
         * Clear a couple of registers. This seems to be
         * necessary when booting via PXE.
         */

	xor	%eax, %eax
	movw	%ax, %ds
	movw	%ax, %es

#ifdef INCLUDE_BIOS_E820_MEM_AUTOSIZE
        .byte   0x67, 0x66
        call    FUNC(vxBiosE820MapQuery)  /* query the BIOS memory map */

#endif /* INCLUDE_BIOS_E820_MEM_AUTOSIZE */

	.byte	0x67, 0x66		/* next inst has 32bit operand */
        lidt	ROM_ADRS(romIdtr)	/* load temporary IDT */

	.byte	0x67, 0x66		/* next inst has 32bit operand */
	lgdt	ROM_ADRS(romGdtr)	/* load temporary GDT */

	/* switch to protected mode */

	mov	%cr0,%eax		/* move CR0 to EAX */
	or	$0x00000001,%eax	/* set the PE bit */
	mov	%eax,%cr0		/* move EAX to CR0 */
	jmp	romInit1		/* near jump to flush a inst queue */

romInit1:
	.byte	0x66			/* next inst has 32bit operand */
	mov	$DATA_SEGMENT,%eax	/* set data segment 0x10 is 3rd one */
	mov	%ax,%ds			/* set DS */
	mov	%ax,%es			/* set ES */
	mov	%ax,%fs			/* set FS */
	mov	%ax,%gs			/* set GS */
	mov	%ax,%ss			/* set SS */
	.byte	0x66			/* next inst has 32bit operand */
	mov	$ ROM_STACK,%esp 	/* set lower mem stack pointer */
	.byte	0x67, 0x66		/* next inst has 32bit operand */
	ljmp	$CODE_SEGMENT, $ ROM_TEXT_ADRS + romInit2 - romInit

	/* cold start code in PROTECTED MODE(32 bits) */

	.balign 16,0x90
romInit2:
	cli				/* LOCK INTERRUPT */
	movl	$ ROM_STACK,%esp	/* set a stack pointer */

	call	FUNC(romA20on)		/* enable A20 */
	cmpl	$0, %eax		/* is A20 enabled? */
	jne	romInitHlt		/*   no: jump romInitHlt */

	movl    $ BOOT_COLD,%ebx	/* %ebx has the startType */
	movl    $ BOOT_COLD,%edi	/* %edi has the startType */

	/* copy bootrom image to dst addr if (romInit != ROM_TEXT_ADRS) */

warm:

#ifdef INCLUDE_BIOS_E820_MEM_AUTOSIZE

	/* copy the BIOS memory map to reserved storage */

	cmpl    $BOOT_COLD, %ebx
	jne     1f
	call	FUNC(vxBiosE820MapCopy)
1:
#endif /* INCLUDE_BIOS_E820_MEM_AUTOSIZE */

	/* Use PAE mode as an indicator that 64 bit mode has been setup */

	movl	%cr4, %eax
	andl	$CR4_PAE, %eax		/* Check whether we are in PAE mode */
	jz	rom_not_PAE

	.code64
	.balign 16

	movl	%edi,%ebx		/* edi has the startType */

	mov	$ STACK_ADRS, %rsp	/* initialise the stack pointer */

	/* Initialize static MMU table */

	mov	$ FUNC(PAGE_PML4_NAME(rom)) - FUNC(romInit), %rax
	add	$ ROM_TEXT_ADRS,%rax
	movq	%rax, %cr3

	jmp	rom_sixtyfour_section

	.code32
	.balign 16

rom_not_PAE:

    xorl    %eax, %eax              /* zero EAX */
    movl    %cr0, %edx              /* get CR0 */
    andl    $0x7ffafff1, %edx       /* clear PG, AM, WP, TS, EM, MP */
    movl    %edx, %cr0              /* set CR0 */

    pushl   %eax                    /* initialize EFLAGS */
    popfl

	xorl	%eax, %eax		/* zero EAX */
	movl	%eax, %cr4		/* initialize CR4 */

	movl	$romGdtr,%eax		/* load the original GDT */
	subl	$ FUNC(romInit),%eax
	addl	$ ROM_TEXT_ADRS,%eax
	pushl	%eax
	call	FUNC(romLoadGdt)

	movl	$ STACK_ADRS, %esp	/* initialise the stack pointer */
	movl    $ ROM_TEXT_ADRS, %esi	/* get src addr(ROM_TEXT_ADRS) */
	movl    $ romInit, %edi		/* get dst addr(romInit) */
	cmpl	%esi, %edi		/* is src and dst same? */
	je	    romInit4		/*   yes: skip copying */
	movl    $ FUNC(end), %ecx	/* get "end" addr */
	subl    %edi, %ecx		/* get nBytes to copy */
	shrl    $2, %ecx		/* get nLongs to copy */
	cld 				/* clear the direction bit */
	rep				/* repeat next inst ECX time */
	movsl                           /* copy itself to the entry point */

romInit4:

	movl    $DATA_SEGMENT,%eax	/* set data segment 0x10 is 3rd one */
	movl    %eax,%ds

	lgdt    ((FUNC(romSysGdtr) - FUNC (romInit)) + ROM_TEXT_ADRS)

	/* load new GDT with the 64bit segments using 32bit descriptor */

	/*
	 * Prepare for entering 64bits mode
	 */

	xorl    %eax, %eax
	orl     $CR4_PAE, %eax          /* PAE */
	movl    %eax, %cr4

	/* Initialize static MMU table */

	movl    $ FUNC(PAGE_PML4_NAME(rom)) - FUNC(romInit), %eax
	addl    $ ROM_TEXT_ADRS,%eax

	movl    %eax, %cr3

	movl    $IA32_EFER, %ecx        /* Extended Feature Enable Register */
	rdmsr

	orl     $IA32_EFER_LME, %eax    /* Enable IA-32e mode */
	wrmsr

	xorl    %eax, %eax
        orl     $CR0_PG, %eax            /* Enable paging */
        orl     $CR0_PE, %eax            /* Enable protected mode */
	movl    %eax, %cr0

        /*
         * IA-32e 32 bit compatibility mode.
         * Jump to 64bit mode using new gdt/idt.
         */

	ljmp    $CODE_SEGMENT, $(romStartMode64 - romInit + ROM_TEXT_ADRS)

	.code64
	.balign 16,0x90

        .globl FUNC(romStartMode64)

FUNC_LABEL(romStartMode64)

	/*
	 * first code that executes in 64 bit mode.
	 */

	movl    %ebx, %esi              /* Save startType */

	movl    $0x80000001, %eax       /* Execute disable check */
	cpuid
	movl    %edx, %edi

	movl    %esi, %ebx              /* Restore startType */

	movl    $IA32_EFER, %ecx        /* Extended Feature Enable Register */
	rdmsr

	orl     $IA32_EFER_SCE, %eax    /* System call enable */

	btl     $20,%edi
	jnc     1f
	orl     $IA32_EFER_NX, %eax     /* Enable execute disable */
1:
	wrmsr

rom_sixtyfour_section:

	mov     $CR0_INIT_VALUE, %rax
	movq    %rax, %cr0

	/* %rsp was already set to $STACK_ADRS, and grows down from there */

	push  $0                        /* initialize EFLAGS */
	popf

	lgdt    ((FUNC(romSysGdtr) - FUNC (romInit)) + ROM_TEXT_ADRS)

        movl    $DATA_SEGMENT,%eax      /* set data segment 0x10 is 3rd one */
        movl    %eax,%ds
        movl    %eax,%ss
        movl    %eax,%es

	/* ebx has the startType, pass it to C routine usrInit() via %edi */

	movl    %ebx, %edi

        /*
         * Run C code in 64 bit address space.
         */

	mov     $ FUNC(romInit), %rax
	push    %rax                    /* fake return address */

	mov     romInitialCcode(%rip), %rax
	jmp     *%rax

	.align  8
	.globl  romInitialCcode
romInitialCcode:
	.quad   romStart

	/* just in case, if there's a problem in romStart or romA20on */

romInitHlt:
	movl    %eax, %edi
	call	FUNC(romEaxShow)	/* show EAX on your display device */
	hlt	
	jmp	romInitHlt

FUNC_END(romInit)
FUNC_END(_romInit)

	.code32

#ifdef INCLUDE_BIOS_E820_MEM_AUTOSIZE

/*
 * MUST include the assembly file at this specific location in
 * romInit.s to ensure correct linkage to 16 bit address space
 */

#include <vxBiosE820ALib.s>

#endif /* INCLUDE_BIOS_E820_MEM_AUTOSIZE */

/*******************************************************************************
*
* romLoadGdt - load the global descriptor table.
*
* RETURNS: N/A
*
* NOMANUAL
*
* void romLoadGdt (char *romGdtr)
*
*/

	.balign 16,0x90
FUNC_LABEL(romLoadGdt)
	movl	SP_ARG1(%esp),%eax
	lgdt	(%eax)
	movw	$DATA_SEGMENT,%ax	/* a selector 0x10 is 3rd one */
	movw	%ax,%ds
	movw	%ax,%es
	movw	%ax,%fs
	movw	%ax,%gs
	movw	%ax,%ss
	ret

/*******************************************************************************
*
* romEaxShow - show EAX register
*
* show EAX register in your display device
*
* SYNOPSIS
* \ss
* void romEaxShow (void)
* \se
*
* RETURNS: N/A
*
*/

	.balign 16,0x90
FUNC_LABEL(romEaxShow)

	/* show EAX register in your display device available */

	ret

/* ************************************************************************ */

romTableStart:
_romTableStart:

	/* temporary IDTR stored in code segment in ROM */

	.balign 16,0x90
romIdtr:
	.word	0x0000			/* size   : 0 */
	.quad	0x00000000		/* address: 0 */

	/* temporary GDTR stored in code segment in ROM */

	.balign 16,0x90
romGdtr:
	.word	0x0027			/* size   : 39(8 * 5 - 1) bytes */
	.quad	(romGdt	- romInit + ROM_TEXT_ADRS) /* address: romGdt */

	/* temporary GDT stored in code segment in ROM */

	.balign 16,0x90
romGdt:
	/* 0(selector=0x0000): Null descriptor */
	.word	0x0000
	.word	0x0000
	.byte	0x00
	.byte	0x00
	.byte	0x00
	.byte	0x00

	/* 1(selector=0x0008): Code descriptor */
	.word	0xffff			/* limit: xffff */
	.word	0x0000			/* base : xxxx0000 */
	.byte	0x00			/* base : xx00xxxx */
	.byte	0x9a			/* Code e/r, Present, DPL0 */
	.byte	0xcf			/* limit: fxxxx, Page Gra, 32bit */
	.byte	0x00			/* base : 00xxxxxx */

	/* 2(selector=0x0010): Data descriptor */
	.word	0xffff			/* limit: xffff */
	.word	0x0000			/* base : xxxx0000 */
	.byte	0x00			/* base : xx00xxxx */
	.byte	0x92			/* Data r/w, Present, DPL0 */
	.byte	0xcf			/* limit: fxxxx, Page Gra, 32bit */
	.byte	0x00			/* base : 00xxxxxx */

	/* 3(selector=0x0018): Code descriptor, for the nesting interrupt */
	.word	0xffff			/* limit: xffff */
	.word	0x0000			/* base : xxxx0000 */
	.byte	0x00			/* base : xx00xxxx */
	.byte	0x9a			/* Code e/r, Present, DPL0 */
	.byte	0xcf			/* limit: fxxxx, Page Gra, 32bit */
	.byte	0x00			/* base : 00xxxxxx */

	/* 4(selector=0x0020): Code descriptor, for the nesting interrupt */
	.word	0xffff			/* limit: xffff */
	.word	0x0000			/* base : xxxx0000 */
	.byte	0x00			/* base : xx00xxxx */
	.byte	0x9a			/* Code e/r, Present, DPL0 */
	.byte	0xcf			/* limit: fxxxx, Page Gra, 32bit */
	.byte	0x00			/* base : 00xxxxxx */

	.balign 16,0x90
romSysGdtr:
	.word   0x0027                  /* size   : 39(8 * 5 - 1) bytes */
	.quad   (romSysGdt - romInit + ROM_TEXT_ADRS)

	.balign 16,0x90
romSysGdt:
	/* 0(selector=0x0000): Null descriptor */
	.word   0x0000
	.word   0x0000
	.byte   0x00
	.byte   0x00
	.byte   0x00
	.byte   0x00

	/* 1(selector=0x0008): Code descriptor, for the supervisor mode task */
	.word   0xffff                  /* limit: xffff */
	.word   0x0000                  /* base : xxxx0000 */
	.byte   0x00                    /* base : xx00xxxx */
	.byte   0x9a                    /* Code e/r, Present, DPL0 */
	.byte   0xaf                    /* limit: fxxxx, Page Gra, 64bit */
	.byte   0x00                    /* base : 00xxxxxx */

	/* 2(selector=0x0010): Data descriptor */
	.word   0xffff                  /* limit: xffff */
	.word   0x0000                  /* base : xxxx0000 */
	.byte   0x00                    /* base : xx00xxxx */
	.byte   0x92                    /* Data r/w, Present, DPL0 */
	.byte   0xcf                    /* limit: fxxxx, Page Gra, 32bit */
	.byte   0x00                    /* base : 00xxxxxx */

	/* 3(selector=0x0018): Code descriptor, for the exception */
	.word   0xffff                  /* limit: xffff */
	.word   0x0000                  /* base : xxxx0000 */
	.byte   0x00                    /* base : xx00xxxx */
	.byte   0x9a                    /* Code e/r, Present, DPL0 */
	.byte   0xaf                    /* limit: fxxxx, Page Gra, 64bit */
	.byte   0x00                    /* base : 00xxxxxx */

	/* 4(selector=0x0020): Code descriptor, for the interrupt */
	.word   0xffff                  /* limit: xffff */
	.word   0x0000                  /* base : xxxx0000 */
	.byte   0x00                    /* base : xx00xxxx */
	.byte   0x9a                    /* Code e/r, Present, DPL0 */
	.byte   0xaf                    /* limit: fxxxx, Page Gra, 64bit */
	.byte   0x00                    /* base : 00xxxxxx */

/*
 * Definition of the static MMU tables for the x86_64 execution mode.
 * This static MMU tables setup the following memory map:
 *
 * Virtual Address     Physical Address    Size (in bytes)     Acces Attributes
 * ----------------------------------------------------------------------------
 * 0x00000000_00000000 0x00000000_00000000 0x00000000_80000000 RWX cached RAM
 * ...
 * 0xffffffff_80000000 0x00000000_00000000 0x00000000_80000000 RWX cached RAM
 *
 * NOTE: if the macro _WRS_BSP_DEBUG_NULL_ACCESS is defined the first page
 *       address range 0x0 to 0xfff is un-mapped to catch NULL pointer read
 *       and write accesses
 */

        /* MMU  entries begin ===================================== */
        PAGE_BEGIN_MMU_VARS

        /* PT   entries begin ------------------------------------- */
#ifdef  _WRS_BSP_DEBUG_NULL_ACCESS
        PAGE_BEGIN_PT_VARS

        PAGE_BEGIN_PT    (rom_null, ROM_TEXT_ADRS)
        PAGE_CREATE_PTS  (1       , STATIC_MMU_TBL_PAGE_ATTR_PROTECTED_4K) /* 000     */
        PAGE_CREATE_PTS  (511     , STATIC_MMU_TBL_PAGE_ATTR_RAM_4K      ) /* 001-511 */
        PAGE_END_NO_FILL
#endif  /* _WRS_BSP_DEBUG_NULL_ACCESS */
        /* PT   entries end --------------------------------------- */

        /* PD   entries begin ------------------------------------- */
        PAGE_BEGIN_PD_VARS

        PAGE_BEGIN_PD    (rom_1to1_1GB, ROM_TEXT_ADRS)
#ifdef  _WRS_BSP_DEBUG_NULL_ACCESS
        PAGE_CREATE_PD   (rom_null, PAGE_P|PAGE_RW|PAGE_US    )   /* 000     */
#else
        PAGE_CREATE_PDS  (1       , STATIC_MMU_TBL_PAGE_ATTR_RAM) /* 000     */
#endif  /* _WRS_BSP_DEBUG_NULL_ACCESS */
        PAGE_CREATE_PDS  (511     , STATIC_MMU_TBL_PAGE_ATTR_RAM) /* 001-511 */
        PAGE_END_NO_FILL

        PAGE_BEGIN_PD    (rom_1to1_2GB, ROM_TEXT_ADRS)
        PAGE_CREATE_PDS  (512     , STATIC_MMU_TBL_PAGE_ATTR_RAM) /* 000-511 */
        PAGE_END_NO_FILL
        /* PD   entries end --------------------------------------- */

        /* PDP  entries begin ------------------------------------- */
        PAGE_BEGIN_PDP   (rom_kernel      , ROM_TEXT_ADRS)
        PAGE_CREATE_PDP  (rom_1to1_1GB    , PAGE_P|PAGE_RW|PAGE_US) /* 000 */
        PAGE_CREATE_PDP  (rom_1to1_2GB    , PAGE_P|PAGE_RW|PAGE_US) /* 001 */
        PAGE_CREATE_BLANK(508)
        PAGE_CREATE_PDP  (rom_1to1_1GB    , PAGE_P|PAGE_RW|PAGE_US) /* 510 */
        PAGE_CREATE_PDP  (rom_1to1_2GB    , PAGE_P|PAGE_RW|PAGE_US) /* 511 */
        PAGE_END
        /* PDP  entries end --------------------------------------- */

        /* PML4 entries begin ------------------------------------- */
	/* Top level of the static MMU translation table. */
        PAGE_PML4_BEGIN  (rom)
        PAGE_CREATE_PML4 (rom_kernel  , PAGE_P|PAGE_RW|PAGE_US) /* 000 */
        PAGE_CREATE_BLANK(510)
        PAGE_CREATE_PML4 (rom_kernel  , PAGE_P|PAGE_RW|PAGE_US) /* 511 */
        PAGE_END
        /* PML4 entries end --------------------------------------- */

        /* MMU  entries end ======================================= */

/* END of static MMU translation table definitions */

romTableEnd:
_romTableEnd:
