/* sysALib.s - system-dependent routines */

/*
 * Copyright (c) 2014-2016 Wind River Systems, Inc.
 *
 * The right to copy, distribute, modify or otherwise make use
 * of this software may be licensed only pursuant to the terms
 * of an applicable Wind River license agreement.
 */

/*
modification history
--------------------
15aug16,emj  fix apigen synopsis tags
09jul15,l_z  add fast warm boot support. (F3387)
09oct15,j_l  Changed second sysInit declaration to _sysInit
19aug15,dot  V7PRO-2102 - fixed apigen errors and warnings.
19mar15,scm  V7PRO-1833 clean up Multiboot/E820 support
30jan15,scm  V7PRO-1656 multibootShow initialization issue 
09sep14,rec  Merge in HV changes from itl_64
13aug14,ten  US44071 - Merge multiboot changes from itl_64 BSP:
                       US38341, V7PRO-891
22apr14,scm  US37149 -sysCalTSC improper register usage...
27mar14,j_l  US30676 - remove config.h
31jan14,j_l  initial creation based on itl_64 version 10jan14
*/

/*
DESCRIPTION
This module contains system-dependent routines written in assembly
language.

This module must be the first specified in the \f3ld\f1 command used to
build the system.  The sysInit() routine is the system start-up code.

INTERNAL
Many routines in this module doesn't use the "c" frame pointer %ebp@ !
This is only for the benefit of the stacktrace facility to allow it 
to properly trace tasks executing within these routines.

SEE ALSO: 
.I "i80386 32-Bit Microprocessor User's Manual"
*/



#define _ASMLANGUAGE
#include <vxWorks.h>
#include <asm.h>
#include <regs.h>
#include <sysLib.h>
#include <pc.h>

#ifdef _WRS_CONFIG_SMP
#include <private/vxSmpP.h>
#endif /* _WRS_CONFIG_SMP */

#ifdef GRUB_MULTIBOOT
#define MMAP_SIZE      0x2c                      /* E820 mem map size offset */
#define MMAP_ADDR      0x30          /* E820 mem map starting address offset */

#define MMAP_PRESENT   0x40                /* bit #6 valid e820 mem map flag */
#endif /* GRUB_MULTIBOOT */

#define SP_ARG1         4

#if defined(INCLUDE_BOOT_APP)
#define SYS_BASE_ADRS   SYS_RAM_HIGH_ADRS
#else
#define SYS_BASE_ADRS   SYS_RAM_LOW_ADRS
#endif

    .data

    /* externals */

    .globl  FUNC(usrInit)                   /* system initialization routine */
#if defined(INCLUDE_WARM_BOOT)
    .globl  FUNC(warmBootBackup)
#endif /* INCLUDE_WARM_BOOT */
    .globl  VAR(sysProcessor)         /* initialized to NONE(-1) in sysLib.c */

#ifdef _WRS_CONFIG_SMP
    _X86_64_ASM_EXTERN_KERNEL_GLOBALS

    .globl  FUNC(sysInitCpuStartup)                       /* the cpu startup */
    .globl  FUNC(sysInitCpuStartupEnd)                    /* the cpu startup */
    .globl  FUNC(cpuInit1)                 /* the cpu startup initialization */
    .globl  FUNC(sysCpuInit)                       /* the cpu initialization */
#endif /* _WRS_CONFIG_SMP */

    /* internals */

    .globl  sysInit                                  /* start of system code */
    .globl  _sysInit                                 /* start of system code */

#ifdef INCLUDE_WARM_BOOT
    .globl  warmBootEntry                            /* start of warm boot   */
#endif

#ifdef GRUB_MULTIBOOT

    /* relocate multiboot header to MULTIBOOT_SCRATCH area */

    .globl  GTEXT(mbhRelocateMH) 
#endif /* GRUB_MULTIBOOT */

    .globl  GTEXT(sysInitGDT)

#if (defined(INCLUDE_BOOT_GO_OVERRIDE)   && \
    defined(BOOT_GO_OVERRIDE_FUNCTION))  || \
    defined(INCLUDE_MULTI_STAGE_BOOT)

    .globl  GTEXT(sysBootElf32)
#endif /* INCLUDE_BOOT_GO_OVERRIDE */

    .globl  GTEXT(sysInByte)
    .globl  GTEXT(sysInWord)
    .globl  GTEXT(sysInLong)

    .globl  GTEXT(sysInWordString)
    .globl  GTEXT(sysInLongString)

    .globl  GTEXT(sysOutByte)
    .globl  GTEXT(sysOutWord)
    .globl  GTEXT(sysOutLong)

    .globl  GTEXT(sysOutWordString)
    .globl  GTEXT(sysOutLongString)

    .globl  GTEXT(sysReboot)
    .globl  GTEXT(sysWait)

    .globl  GTEXT(sysLoadGdt)
    .globl  GTEXT(sysGdtr)
    .globl  GTEXT(sysGdt)

#if (defined(INCLUDE_BOOT_GO_OVERRIDE)   && \
    defined(BOOT_GO_OVERRIDE_FUNCTION))  || \
    defined(INCLUDE_MULTI_STAGE_BOOT)    || \
    defined(_WRS_CONFIG_SMP) 

    .globl  GTEXT(sysGdt32)
#endif /* INCLUDE_BOOT_GO_OVERRIDE */

    .globl  sysTableStart
    .globl  _sysTableStart
    .globl  sysTableEnd
    .globl  _sysTableEnd

    .globl  GDATA(sysCsSuper)              /* code selector: supervisor mode */
    .globl  GDATA(sysCsExc)                      /* code selector: exception */
    .globl  GDATA(sysCsInt)                      /* code selector: interrupt */
    .globl  GDATA(sysSsSuper)             /* Stack selector: supervisor mode */

    .globl  GTEXT(sysCalTSC)

    /* locals */

#ifdef INCLUDE_HYPERVISOR
	.globl VAR(hvHeap)		/* hypervisor heap storage */
	.globl VAR(hvHeapSize)
#endif /* INCLUDE_HYPERVISOR */

	/* locals */

FUNC_LABEL(vendorIdIntel)		          /* CPUID vendor ID - Intel */
    .ascii  "GenuineIntel"

/*
 * Create the text section to be placed in low memory. Set the attributes 
 * explicitly because this is not the standard .text section (for which
 * the attributes are set by default). The attributes required are:
 * CONTENTS, ALLOC, LOAD, READONLY, CODE
 */

    .section .text_locore,"ax"
    .code32
    .balign 16

/*******************************************************************************
*
* sysInit - start after boot
*
* This routine is the system start-up entry point for VxWorks in RAM, the
* first code executed after booting.  It disables interrupts, sets up
* the stack, and jumps to the C routine usrInit() in usrConfig.c.
*
* The initial stack is set to grow down from the address of sysInit().  This
* stack is used only by usrInit() and is never used again.  Memory for the
* stack must be accounted for when determining the system load address.
*
* NOTE: This routine should not be called by the user.
*
* SYNOPSIS
* \ss
* sysInit ()              /@ THIS IS NOT A CALLABLE ROUTINE @/
* \se
*
*
* RETURNS: N/A
*/

sysInit:
_sysInit:

    /*
     * We're coming here from the boot loader code, and the CPU is
     * already running in 64-bit execution mode.
     */

    cli                         /* LOCK INTERRUPT */

#if    defined(INCLUDE_WARM_BOOT)
    /*
     * Overload startType in case bootApp has done a warm reboot.
     * If this is a kernel warm reboot the entry point will be
     * right afterwards and startType will be BOOT_NORMAL.
     */

    movl    $BOOT_COLD, %edi

warmBootEntry:
    cli
#endif /* INCLUDE_WARM_BOOT */

    /* Use PAE mode as an indicator that 64 bit mode has been setup */
      
    movl    %cr4, %esi
    andl    $CR4_PAE, %esi          /* Check whether we are in PAE mode */
    jz      not_PAE

    .code64
    .balign 16,0x90

    movl    %edi,%ebx               /* edi has the startType */
        
#if    defined(INCLUDE_BOOT_RAM_IMAGE)
    mov	    $ RAM_HIGH_ADRS,%rsp	/* initialize stack pointer */
#elif !defined(INCLUDE_BOOT_APP)
    mov     $ RAM_LOW_ADRS,%rsp	/* initialize stack pointer */
#endif

    /* Initialize static MMU table */

    mov     $ FUNC(PAGE_PML4_NAME(sys)), %rax
    movq    %rax, %cr3

    jmp     sixtyfour_section

    .code32
    .balign 16,0x90
        
not_PAE:
        
#ifdef GRUB_MULTIBOOT
    jmp     multiboot_entry

    /* Multiboot Header ---allows GRUB to load kernel */
    .align 8

    /* MULTIBOOT ---ELF images */
multiboot_header:
    .long   MULTIBOOT_HEADER_MAGIC
    .long   MULTIBOOT_HEADER_FLAGS_ELF
    .long   -(MULTIBOOT_HEADER_MAGIC + MULTIBOOT_HEADER_FLAGS_ELF)

multiboot_entry:

    cmp     $MULTIBOOT_BOOTLOADER_MAGIC, %eax
    jne     multiPrologEnd

    movl    %eax,(MULTIBOOT_SCRATCH)                            /* save magic */

    movl    %ebx,(MULTIBOOT_SCRATCH + 8)           /* save multi-boot pointer */

    /*
     * Upon entry, the EBX register contains the physical address of a Multiboot
     * information data structure, through which the boot loader communicates 
     * vital information. We can use or ignore any parts of the structure as we 
     * choose.
     *
     * The Multiboot information structure and its related substructures may be 
     * placed anywhere in memory by the boot loader (with the exception of the 
     * memory reserved for the kernel and boot modules). We need to avoid 
     * overwriting this memory until we are done using it.
     *
     * The format of the Multiboot information structure (as defined so far) 
     * follows:
     * 
     *           +-------------------+
     *   0       | flags             |    (required)
     *           +-------------------+
     *   4       | mem_lower         |    (present if flags[0] is set)
     *   8       | mem_upper         |    (present if flags[0] is set)
     *           +-------------------+
     *   12      | boot_device       |    (present if flags[1] is set)
     *           +-------------------+
     *   16      | cmdline           |    (present if flags[2] is set)
     *           +-------------------+
     *   20      | mods_count        |    (present if flags[3] is set)
     *   24      | mods_addr         |    (present if flags[3] is set)
     *           +-------------------+
     *   28 - 40 | syms              |    (present if flags[4] or
     *           |                   |                flags[5] is set)
     *           +-------------------+
     *   44      | mmap_length       |    (present if flags[6] is set)
     *   48      | mmap_addr         |    (present if flags[6] is set)
     *           +-------------------+
     *   52      | drives_length     |    (present if flags[7] is set)
     *   56      | drives_addr       |    (present if flags[7] is set)
     *           +-------------------+
     *   60      | config_table      |    (present if flags[8] is set)
     *           +-------------------+
     *   64      | boot_loader_name  |    (present if flags[9] is set)
     *           +-------------------+
     *   68      | apm_table         |    (present if flags[10] is set)
     *           +-------------------+
     *   72      | vbe_control_info  |    (present if flags[11] is set)
     *   76      | vbe_mode_info     |
     *   80      | vbe_mode          |
     *   82      | vbe_interface_seg |
     *   84      | vbe_interface_off |
     *   86      | vbe_interface_len |
     *           +-------------------+
     *
     * The first long word indicates the presence and validity of other fields 
     * in the Multiboot information structure. All as-yet-undefined bits must be
     * set to zero by the boot loader. Any set bits that the operating system 
     * does not understand should be ignored. 
     *
     * If bit 6 in the flags word is set, then the 'mmap_*' fields are valid, 
     * and indicate the address and length of a buffer containing a memory map 
     * of the machine provided by the bios. 
     *
     * The map provided is guaranteed to list all standard ram that should be 
     * available for normal use.
     */

    pushl   %eax                       /* save off MULTIBOOT_BOOTLOADER_MAGIC */

    movl    (%ebx),%eax
    andl    $MMAP_PRESENT,%eax

    cmpl    $MMAP_PRESENT,%eax       /* check if 'mmap_*' is considered valid */
    jne     mbh_mmap_not_valid 

    pushl   %ebx             /* push address of MB information data structure */
    call    FUNC(mbhRelocateMH)             /* relocate multiboot header info */
    popl    %ebx              /* pop address of MB information data structure */

mbh_mmap_not_valid:

    popl    %eax                        /* restore MULTIBOOT_BOOTLOADER_MAGIC */

multiPrologEnd:
    movl    $BOOT_CLEAR,%ebx
#else
    movl    SP_ARG1(%esp),%ebx
#endif /* GRUB_MULTIBOOT */

    movl    $SYS_BASE_ADRS,%esp	/* initialize stack pointer */

    movl    $0,%ebp                 /* initialize frame pointer */

    xorl    %eax, %eax              /* zero EAX */
    movl    %cr0, %edx              /* get CR0 */
    andl    $0x7ffafff1, %edx       /* clear PG, AM, WP, TS, EM, MP */
    movl    %edx, %cr0              /* set CR0 */

    pushl   %eax                    /* initialize EFLAGS */
    popfl

    xorl    %eax, %eax              /* zero EAX */
    movl    %eax, %cr4              /* initialize CR4 */


    /* load new GDT with the 64bit segments using 32bit descriptor */

    lgdt    (FUNC(sysGdtr))

    /*
     * Initialize the %ds segment register
     * We must do this _after_ loading the new GDT, not before.
     */

    movw    $DATA_SEGMENT, %ax	/* set data segment 0x10 is 3rd one */
    movw    %ax,%ds

    /*
     * Prepare for entering 64bits mode
     */

    xorl    %eax, %eax
    orl     $CR4_PAE, %eax          /* PAE */
    movl    %eax, %cr4

    /* Initialize static MMU table */

    movl    $ FUNC(PAGE_PML4_NAME(sys)), %eax

    mov     %eax, %cr3

    movl    $IA32_EFER, %ecx        /* Extended Feature Enable Register */
    rdmsr
        
    orl     $IA32_EFER_LME, %eax    /* Enable IA-32e mode */
    wrmsr

    xorl    %eax, %eax
    orl     $CR0_PG, %eax           /* Enable paging */
    orl     $CR0_PE, %eax           /* Enable protected mode */
    movl    %eax, %cr0

    /*
     * In IA-32e, 32bit compatibility mode.  
     * We now want to jump into 64bit mode, we use the new
     * gdt/idt to do so.
     */

    ljmp    $CODE_SEGMENT, $(sysInitMode64)

#ifdef GRUB_MULTIBOOT
/*******************************************************************************
*
* mbhRelocateMH - relocate multiboot header to MULTIBOOT_SCRATCH area.
*
* RETURNS: N/A
*
* NOMANUAL
*
* void mbhRelocateMH (unsigned long mbHAddr)
*
* NOTE: 
*
* %EBX register is expected to contain the physical address of a Multiboot 
* header information data structure.
*
*/
     
    .balign 16,0x90
FUNC_LABEL(mbhRelocateMH)

    /* 
     * On entry %ebx should hold addr of MB information data structure, 
     * passing on stack...
     */

    movl   SP_ARG1(%esp),%ebx 
 
    /* relocate multiboot header to be visible in 64 bit address space */
                                            
    pushfl                                                     /* save EFLAGS */
    cli                                                     /* LOCK INTERRUPT */
                                            
    /* adjust table pointers prior to relocation to MULTIBOOT_SCRATCH */
                                            
    movl    $MULTIBOOT_SCRATCH, %edi        
    addl    $0x10, %edi                                   /* header dest addr */
                                            
    movl    MMAP_ADDR(%ebx),%edx               
    subl    %ebx, %edx                      
    addl    %edi, %edx                      
                                            
    movl    %ebx,%eax                             /* get header location %esi */
    addl    $MMAP_ADDR, %eax                
    movl    %edx,(%eax)                     
                                            
    movl    %edx, %ecx /* calculate table size up to and including memory map */
    movl    MMAP_SIZE(%ebx),%edx               /* retrieve size of memory map */
    addl    %edx, %ecx                                      /* ending address */
    subl    %edi, %ecx                                  /* less start address */
    shrl    $2, %ecx                                    /* get nLongs to copy */

    /* relocation of multiboot header to MULTIBOOT_SCRATCH */
                                            
    movl    %ebx,%esi                             /* get header location %esi */
                                            
    movl    $MULTIBOOT_SCRATCH, %edi        
    addl    $0x10, %edi                                   /* header dest addr */
                                            
    movl    %edi,(MULTIBOOT_SCRATCH + 8) /* update save of multi-boot pointer */
                                            
    rep                                         /* repeat next inst ECX times */
    movsl                                                  /* copy table data */
 
    popfl                                                   /* restore EFLAGS */

    ret
#endif /* GRUB_MULTIBOOT */

#if (defined(INCLUDE_BOOT_GO_OVERRIDE)   && \
     defined(BOOT_GO_OVERRIDE_FUNCTION)) || \
     defined(INCLUDE_MULTI_STAGE_BOOT) 
/*******************************************************************************
*
* sysInit32 - Start 32 bit vxWorks image
*
* SYNOPSIS
* \ss
* void sysInit32 (entryPoint, startType)
*     UINT32 entryPoint;	/@ entry point address @/
*     int    startType;		/@ system start type @/
* \se
*
*/
        .balign 16,0x90
        .globl FUNC(sysInit32)
FUNC_LABEL(sysInit32)
        /* In 32 bit compatibility mode */
        
        mov     $ FUNC(PAGE_PML4_NAME(sys)), %eax
        mov     %eax, %cr3

        jmp     movToL4Static32          /* flush prefetch queue */
movToL4Static32:

        /* Deactivate IA-32e mode */
        mov     %cr0,%eax
        and     $(~CR0_PG),%eax
        mov     %eax,%cr0

        /* Disable IA-32e mode */
        movl    $IA32_EFER, %ecx
        rdmsr
        and     $(~IA32_EFER_LME),%eax
        wrmsr

        xorl    %eax, %eax
        movl    %cr0, %edx
        andl    $(~(CR0_PG|CR0_AM|CR0_WP|CR0_TS|CR0_EM|CR0_MP)), %edx
        movl    %edx, %cr0
        
        pop     %eax                    /* entryPoint */
        pop     %ebx                    /* startType */
        
        xorl    %ebp, %ebp              /* initialize the frame pointer */
        pushl   $0                      /* initialise the EFLAGS */
        popfl
        pushl   %ebx                    /* push the startType */
        call    *%eax                   /* go to entryPoint */
        
        /* just in case, if there's a problem */

sysInit32Hlt:
        movl    %eax, %edi
        hlt
        jmp     sysInit32Hlt

FUNC_LABEL(sysInit32End)
#endif /* INCLUDE_BOOT_GO_OVERRIDE */

#ifdef _WRS_CONFIG_SMP
/*******************************************************************************
*
* sysInitCpuStartup - istartup initialization for APs...
*
* SYNOPSIS
* \ss
* void sysInitCpuStartup (void)
*
* scratch memory usage:
*
*    scratchMem (scratch memory offset)  scratchPoint
*
*    Standard GDT Entries:
*
*    null descriptor                     scratchMem + 0x04
*
*    kernel text descriptor              scratchMem + 0x0c
*
*    kernel data descriptor              scratchMem + 0x14
*
*    exception text descriptor           scratchMem + 0x1C
*
*    interrupt data descriptor           scratchMem + 0x24
*
*    gdt limit << 16                     scratchMem + LIM_GDT_OFFSET
*    gdt base                            scratchMem + BASE_GDT_OFFSET
*    address of page directory           scratchMem + CR3_OFFSET
*
*    idt limit                           scratchMem + LIM_IDT_OFFSET
*    idt base address                    scratchMem + BASE_IDT_OFFSET
*
*    initial AP stack addr               scratchMem + AP_STK_OFFSET
* \se
*
*/
        .balign 16,0x90
FUNC_LABEL(sysInitCpuStartup)
                                                                         
        /* cold start code in REAL MODE (16 bits) */
                                                                          
        cli                              /* LOCK INTERRUPT */
                                                                         
	xor	%eax, %eax
        
	movw	%ax, %ds
	movw	%ax, %es

        .byte   0x67, 0x66               /* next inst has 32bit operand */
        mov     $CPU_SCRATCH_POINT, %eax /* find the location of scratch mem */
	.byte	0x67, 0x66		/* next inst has 32bit operand */
        mov     (%eax), %eax
        mov     %ax, %sp

        /* load intial GDT */
                                                                          
        .byte   0x67, 0x66               /* next inst has 32bit operand */
        mov     $CPU_SCRATCH_POINT, %eax /* find the location of scratch mem */
	.byte	0x67, 0x66		/* next inst has 32bit operand */
        mov     (%eax), %eax
        .byte   0x67, 0x66               /* next inst has 32bit operand */
        add     $CPU_SCRATCH_GDT, %eax   /* base of initial gdt */
        
        .byte   0x67, 0x66               /* next inst has 32bit operand */
        lgdt    (%eax)
        
	/* switch to protected mode */

	mov	%cr0,%eax		/* move CR0 to EAX */
	or	$0x00000001,%eax	/* set the PE bit */
	mov	%eax,%cr0		/* move EAX to CR0 */
        
	jmp	clearSegs		/* near jump to flush a inst queue */

clearSegs:
                                                                        
        .byte   0x66                    /* next inst has 32bit operand */
	mov	$DATA_SEGMENT,%eax	/* set data segment 0x10 is 3rd one */
	mov	%ax,%ds			/* set DS */
	mov	%ax,%es			/* set ES */
	mov	%ax,%fs			/* set FS */
	mov	%ax,%gs			/* set GS */
	mov	%ax,%ss			/* set SS */
        
        /* jump to protected mode code in high mem */
                                                                        
        .byte   0x67, 0x66               /* next inst has 32bit operand */
	ljmp	$CODE_SEGMENT, $cpuInit1

FUNC_LABEL(sysInitCpuStartupEnd)
FUNC_LABEL(_sysInitCpuStartupEnd)
        
	.balign 16,0x90
FUNC_LABEL(cpuInit1)

	/* cold start code in PROTECTED MODE(32 bits) */

	cli				/* LOCK INTERRUPT */
        
	movl	$0,%ebp			/* initialize frame pointer */

        xorl    %eax, %eax              /* zero EAX */
        movl    %cr0, %edx              /* get CR0 */
        andl    $0x7ffafff1, %edx       /* clear PG, AM, WP, TS, EM, MP */
        movl    %edx, %cr0              /* set CR0 */

        pushl   %eax                    /* initialize EFLAGS */
        popfl 

	xorl	%eax, %eax		/* zero EAX */
	movl	%eax, %cr4		/* initialize CR4 */

        /* Initialize the %ds segment register */

        movl    $DATA_SEGMENT, %eax	/* set data segment 0x10 is 3rd one */
        movl    %eax,%ds

        /* load new GDT with the 64bit segments using 32bit descriptor */

        lgdt    (FUNC(sysGdtr))

	/*
	 * Prepare for entering 64bits mode
	 */

        movl    $0x000006D0,%eax /* set the OSXMMEXCPT,OSFXSR,PGE,PSE,MCE bits */
	orl     $CR4_PAE, %eax          /* PAE */
	movl    %eax, %cr4

	/* Initialize MMU table */

        mov     $CPU_SCRATCH_POINT, %eax /* find the location of scratch mem */
        mov     (%eax), %eax
        add     $CR3_OFFSET, %eax        /* stored value for cr3 */
                                                                          
        mov     (%eax), %eax
        mov     %eax, %cr3

	movl    $IA32_EFER, %ecx        /* Extended Feature Enable Register */
	rdmsr

	orl     $IA32_EFER_LME, %eax    /* Enable IA-32e mode */
	wrmsr

	xorl    %eax, %eax
        orl     $CR0_PG, %eax            /* Enable paging */
        orl     $CR0_PE, %eax            /* Enable protected mode */
	movl    %eax, %cr0

        /*
         * In IA-32e, 32bit compatibility mode.  
         * We now want to jump into 64bit mode, we use the new
         * gdt/idt to do so.
         */

        ljmp    $CODE_SEGMENT, $sysInitCpuStartupMode64

	.code64
	.balign 16,0x90
        
        .globl FUNC(sysInitCpuStartupMode64)
        
FUNC_LABEL(sysInitCpuStartupMode64)

	movl    $0x80000001, %eax       /* Execute disable check */
	cpuid
	movl    %edx, %edi

	movl    $IA32_EFER, %ecx        /* Extended Feature Enable Register */
	rdmsr

	orl     $IA32_EFER_SCE, %eax    /* System call enable */

	btl     $20,%edi
	jnc     1f
	orl     $IA32_EFER_NX, %eax     /* Enable execute disable */
1:
	wrmsr

	mov     $CR0_INIT_VALUE, %rax
	movq    %rax, %cr0

	/* %rsp was already set to $STACK_ADRS, and grows down from there */

	push  $0                        /* initialize EFLAGS */
	popf

        lgdt    (FUNC(sysGdtr))
 
        movl $DATA_SEGMENT,%eax		/* set data segment 0x10 is 3rd one */
        movl %eax,%ds
        movl %eax,%ss
        movl %eax,%es

        /* load intial IDT */

        mov     $CPU_SCRATCH_POINT, %rax /* find the location of scratch mem */
        mov     (%rax), %rax
        add     $CPU_SCRATCH_IDT, %rax   /* base of initial idt */

        lidt    (%rax)

	mov	$CPU_SCRATCH_POINT, %eax
	mov	(%eax), %eax
	add	$AP_STK_OFFSET, %eax
	movq	(%eax), %rsp

        /*
         * Run C code in 64 bit address space.
         */

	mov     $ FUNC(sysInitCpuStartupMode64), %rax
	push    %rax                    /* fake return address */

        mov     cpuInitialCcode(%rip),%rax
        jmp     *%rax

        .align  8
        .globl  cpuInitialCcode
cpuInitialCcode:
        .quad   sysCpuInit

        /* just in case there's a problem */

cpuInitHlt1:
        hlt
        jmp     cpuInitHlt1
#endif /* _WRS_CONFIG_SMP */

        .code64
        .balign 16,0x90

        .globl FUNC(sysInitMode64)

FUNC_LABEL(sysInitMode64)

        movl    %ebx, %esi              /* Save startType */

        movl    $0x80000001, %eax       /* Execute disable check */
        cpuid
        movl    %edx, %edi

        movl    %esi, %ebx              /* Restore startType */

        movl    $IA32_EFER, %ecx        /* Extended Feature Enable Register */
        rdmsr

        orl     $IA32_EFER_SCE, %eax    /* System call enable */

        btl     $20,%edi
        jnc     1f
        orl     $IA32_EFER_NX, %eax     /* Enable execute disable */
1:
        wrmsr

sixtyfour_section:      

        mov     $CR0_INIT_VALUE, %rax
        movq    %rax, %cr0

        /* %rsp was already set to $STACK_ADRS, and grows down from there */

        push  $0                        /* initialize EFLAGS */
        popf

        lgdt    (FUNC(sysGdtr))

        movw $DATA_SEGMENT,%ax		/* set data segment 0x10 is 3rd one */
        movw %ax,%ds
        movw %ax,%ss
        movw %ax,%es

        /* ebx has the startType, pass it to C routine usrInit() via %edi */

        movl    %ebx, %edi

       /* initiate backup of warm boot data if startType is BOOT_COLD */

#if defined (INCLUDE_WARM_BOOT)
        xor     %rdi, %rdi 
        cmpl    $BOOT_COLD, %ebx
        jne     warmBoot_skip

        call    warmBootBackup

warmBoot_skip:
#endif /* INCLUDE_WARM_BOOT */

        /*
         * Run C code in 64 bit address space.
         */

        mov     $ FUNC(sysInitMode64), %rax
        push    %rax                    /* fake return address */

	/* Start the hypervisor if it is linked in (otherwise calls stub) */

#ifdef INCLUDE_HYPERVISOR
        mov     hvInitCode(%rip),%rax
        call    *%rax
#endif

        mov     initialCcode(%rip),%rax
        jmp     *%rax

        .align  8
        .globl  initialCcode
initialCcode:
        .quad   usrInit
#ifdef INCLUDE_HYPERVISOR
hvInitCode:
        .quad   sysEnterHypervisor64
#endif

        /* just in case, if there's a problem */

sysInitHlt:
        movl    %eax, %edi
        hlt
        jmp     sysInitHlt

FUNC_END(sysInit)
FUNC_END(_sysInit)

/* Create the data section to be placed in low memory. Set the attributes 
 * explicitly because this is not the standard .data section (for which
 * the attributes are set by default). The attributes required are:
 * CONTENTS, ALLOC, LOAD, DATA
 */

	.section .data_locore,"aw"

/*
 * Allocation of the memory area which the hypervisor kernel will use for
 * its main memory storage.
 */

#ifdef INCLUDE_HYPERVISOR

	.section .hvbss,"aw",@nobits

hvHeap:
	.skip	HYPERVISOR_RAM_SIZE

	.section .data_locore,"aw"

hvHeapSize:
	.quad	HYPERVISOR_RAM_SIZE

#endif /* INCLUDE_HYPERVISOR */


/**************************************************************************/

sysTableStart:
_sysTableStart:

#if (defined(INCLUDE_BOOT_GO_OVERRIDE)   && \
     defined(BOOT_GO_OVERRIDE_FUNCTION)) || \
    defined(INCLUDE_MULTI_STAGE_BOOT) 
        .balign 16,0x90
sysGdtr32:
        .word   0x0027
        .quad   (sysGdt32)
#endif /* INCLUDE_BOOT_GO_OVERRIDE */
        
#if (defined(INCLUDE_BOOT_GO_OVERRIDE)   && \
     defined(BOOT_GO_OVERRIDE_FUNCTION)) || \
    defined(INCLUDE_MULTI_STAGE_BOOT)    || \
    defined(_WRS_CONFIG_SMP)
        .balign 16,0x90
sysGdt32:
        /* 0(selector=0x0000): Null descriptor */
        .word   0x0000
        .word   0x0000
        .byte   0x00
        .byte   0x00
        .byte   0x00
        .byte   0x00

        /* 1(selector=0x0008): Code descriptor */
        .word   0xffff                  /* limit: xffff */
        .word   0x0000                  /* base : xxxx0000 */
        .byte   0x00                    /* base : xx00xxxx */
        .byte   0x9a                    /* Code e/r, Present, DPL0 */
        .byte   0xcf                    /* limit: fxxxx, Page Gra, 32bit */
        .byte   0x00                    /* base : 00xxxxxx */

        /* 2(selector=0x0010): Data descriptor */
        .word   0xffff                  /* limit: xffff */
        .word   0x0000                  /* base : xxxx0000 */
        .byte   0x00                    /* base : xx00xxxx */
        .byte   0x92                    /* Data r/w, Present, DPL0 */
        .byte   0xcf                    /* limit: fxxxx, Page Gra, 32bit */
        .byte   0x00                    /* base : 00xxxxxx */

        /* 3(selector=0x0018): Code descriptor, for the nesting interrupt */
        .word   0xffff                  /* limit: xffff */
        .word   0x0000                  /* base : xxxx0000 */
        .byte   0x00                    /* base : xx00xxxx */
        .byte   0x9a                    /* Code e/r, Present, DPL0 */
        .byte   0xcf                    /* limit: fxxxx, Page Gra, 32bit */
        .byte   0x00                    /* base : 00xxxxxx */

        /* 4(selector=0x0020): Code descriptor, for the nesting interrupt */
        .word   0xffff                  /* limit: xffff */
        .word   0x0000                  /* base : xxxx0000 */
        .byte   0x00                    /* base : xx00xxxx */
        .byte   0x9a                    /* Code e/r, Present, DPL0 */
        .byte   0xcf                    /* limit: fxxxx, Page Gra, 32bit */
        .byte   0x00                    /* base : 00xxxxxx */
#endif /* INCLUDE_BOOT_GO_OVERRIDE */

        .balign 16,0x90
FUNC_LABEL(sysGdtr)
        .word   0x0027                  /* size   : 39(8 * 5 - 1) bytes */
        .quad   sysGdt

        .balign 16,0x90
FUNC_LABEL(sysGdt)
        /* 0(selector=0x0000): Null descriptor */
        .word   0x0000
        .word   0x0000
        .byte   0x00
        .byte   0x00
        .byte   0x00
        .byte   0x00

        /* 1(selector=0x0008): Code descriptor, for the supervisor mode task */
        .word   0xffff                  /* limit: xffff */
        .word   0x0000                  /* base : xxxx0000 */
        .byte   0x00                    /* base : xx00xxxx */
        .byte   0x9a                    /* Code e/r, Present, DPL0 */
        .byte   0xaf                    /* limit: fxxxx, Page Gra, 64bit */
        .byte   0x00                    /* base : 00xxxxxx */

        /* 2(selector=0x0010): Data descriptor */
        .word   0xffff                  /* limit: xffff */
        .word   0x0000                  /* base : xxxx0000 */
        .byte   0x00                    /* base : xx00xxxx */
        .byte   0x92                    /* Data r/w, Present, DPL0 */
        .byte   0xcf                    /* limit: fxxxx, Page Gra, 32bit */
        .byte   0x00                    /* base : 00xxxxxx */

        /* 3(selector=0x0018): Code descriptor, for the exception */
        .word   0xffff                  /* limit: xffff */
        .word   0x0000                  /* base : xxxx0000 */
        .byte   0x00                    /* base : xx00xxxx */
        .byte   0x9a                    /* Code e/r, Present, DPL0 */
        .byte   0xaf                    /* limit: fxxxx, Page Gra, 64bit */
        .byte   0x00                    /* base : 00xxxxxx */

        /* 4(selector=0x0020): Code descriptor, for the interrupt */
        .word   0xffff                  /* limit: xffff */
        .word   0x0000                  /* base : xxxx0000 */
        .byte   0x00                    /* base : xx00xxxx */
        .byte   0x9a                    /* Code e/r, Present, DPL0 */
        .byte   0xaf                    /* limit: fxxxx, Page Gra, 64bit */
        .byte   0x00                    /* base : 00xxxxxx */

/*
 * Definition of the static MMU tables for the x86_64 execution mode.
 * This static MMU tables setup the following memory map:
 *
 * Virtual Address     Physical Address    Size (in bytes)     Acces Attributes
 * ----------------------------------------------------------------------------
 * 0x00000000_00000000 0x00000000_00000000 0x00000000_80000000 RWX cached RAM
 * ...
 * 0xffffffff_80000000 0x00000000_00000000 0x00000000_80000000 RWX cached RAM
 *
 * NOTE: if the macro _WRS_BSP_DEBUG_NULL_ACCESS is defined the first page
 *       address range 0x0 to 0xfff is un-mapped to catch NULL pointer read
 *       and write accesses
 */

        /* MMU  entries begin ===================================== */
        PAGE_BEGIN_MMU_VARS
        
        /* PT   entries begin ------------------------------------- */
#ifdef  _WRS_BSP_DEBUG_NULL_ACCESS
        PAGE_BEGIN_PT_VARS

        PAGE_BEGIN_PT    (sys_null, SYS_BASE_ADRS)
        PAGE_CREATE_PTS  (1       , STATIC_MMU_TBL_PAGE_ATTR_PROTECTED_4K) /* 000     */
        PAGE_CREATE_PTS  (511     , STATIC_MMU_TBL_PAGE_ATTR_RAM_4K      ) /* 001-511 */
        PAGE_END_NO_FILL
#endif  /* _WRS_BSP_DEBUG_NULL_ACCESS */
        /* PT   entries end --------------------------------------- */
        
        /* PD   entries begin ------------------------------------- */
        PAGE_BEGIN_PD_VARS
        
        PAGE_BEGIN_PD    (sys_1to1_1GB, SYS_BASE_ADRS)
#ifdef  _WRS_BSP_DEBUG_NULL_ACCESS
        PAGE_CREATE_PD   (sys_null, PAGE_P|PAGE_RW|PAGE_US    )   /* 000     */
#else
        PAGE_CREATE_PDS  (1       , STATIC_MMU_TBL_PAGE_ATTR_RAM) /* 000     */
#endif  /* _WRS_BSP_DEBUG_NULL_ACCESS */
        PAGE_CREATE_PDS  (511     , STATIC_MMU_TBL_PAGE_ATTR_RAM) /* 001-511 */
        PAGE_END_NO_FILL

        PAGE_BEGIN_PD    (sys_1to1_2GB, SYS_BASE_ADRS)
        PAGE_CREATE_PDS  (512     , STATIC_MMU_TBL_PAGE_ATTR_RAM) /* 000-511 */
        PAGE_END_NO_FILL
        /* PD   entries end --------------------------------------- */
        
        /* PDP  entries begin ------------------------------------- */
        PAGE_BEGIN_PDP   (sys_kernel      , SYS_BASE_ADRS)
        PAGE_CREATE_PDP  (sys_1to1_1GB    , PAGE_P|PAGE_RW|PAGE_US) /* 000 */
        PAGE_CREATE_PDP  (sys_1to1_2GB    , PAGE_P|PAGE_RW|PAGE_US) /* 001 */
        PAGE_CREATE_BLANK(508)
        PAGE_CREATE_PDP  (sys_1to1_1GB    , PAGE_P|PAGE_RW|PAGE_US) /* 510 */
        PAGE_CREATE_PDP  (sys_1to1_2GB    , PAGE_P|PAGE_RW|PAGE_US) /* 511 */
        PAGE_END
        /* PDP  entries end --------------------------------------- */
        
        /* PML4 entries begin ------------------------------------- */
	/* Top level of the static MMU translation table. */
        PAGE_PML4_BEGIN  (sys)
        PAGE_CREATE_PML4 (sys_kernel  , PAGE_P|PAGE_RW|PAGE_US) /* 000 */
        PAGE_CREATE_BLANK(510)
        PAGE_CREATE_PML4 (sys_kernel  , PAGE_P|PAGE_RW|PAGE_US) /* 511 */
        PAGE_END
        /* PML4 entries end --------------------------------------- */

        /* MMU  entries end ======================================= */
        
/* END of static MMU translation table definitions */

sysTableEnd:
_sysTableEnd:

	.text

#if (defined(INCLUDE_BOOT_GO_OVERRIDE)   && \
     defined(BOOT_GO_OVERRIDE_FUNCTION)) || \
    defined(INCLUDE_MULTI_STAGE_BOOT) 
/*******************************************************************************
*
* sysBootElf32 - Prepare for 32 bit vxWorks image
*
* SYNOPSIS
* \ss
* void sysBootElf32 (entryPoint, startType)
*     UINT32 entryPoint;	/@ entry point address @/
*     int    startType;		/@ system start type @/
* \se
*
*/
        .balign 16,0x90
FUNC_LABEL(sysBootElf32)
	/* Initialize static MMU table */
        mov     $ FUNC(PAGE_PML4_NAME(sys)), %rax
        mov     %rax, %cr3
        jmp     movToL4Static            /* flush prefetch queue */
movToL4Static:

        /*
         * Prepare for entering 32 bit compatibility mode
         */

        /* switch to GDT for static map */
        lgdt    (FUNC(sysGdtr32))

        /* Save entryPoint and startType */
        mov     %rdi, %rbx
        shl     $32, %rsi
        or      %rsi, %rbx
        
        /* Copy stub code to WARMBOOT_COPY_ADRS */
        cld
        mov     $ (FUNC(sysInit32) + ~LOCAL_MEM_PHYS_ADRS_MASK),%rsi  /* src addr */
        mov     $ WARMBOOT_COPY_ADRS,%rdi /* dst addr */
        mov     $ FUNC(sysInit32End)-FUNC(sysInit32),%rcx /* size to copy */
        shr     $2,%rcx                 /* get nLongs to copy */
        rep                             /* repeat next inst ECX time */
        movsl                           /* copy nLongs from src to dst */
        
        /* adjust stack pointer for 32 bit use */
        mov     $SYS_BASE_ADRS, %rsp

        /* force return to IA-32e compatability mode ? */

        pushq   %rbx                    /* entryPoint and startType */
        pushq   $CODE_SEGMENT
        movq    %rax,%rbp
        pushq   $WARMBOOT_COPY_ADRS     /* sysInit32 */
        lretq
#endif /* INCLUDE_BOOT_GO_OVERRIDE */
        
/*******************************************************************************
*
* sysInitGDT - load the brand new GDT into RAM
*
* SYNOPSIS
* \ss
* void sysInitGDT (void)
* \se
*
*/
        .balign 16,0x90
FUNC_LABEL(sysInitGDT)

        pushal

        /* copy GTP from sysGdt to *pSysGdt */

        mov     %rsp,%rbp

        mov     $FUNC(sysGdt),%rsi      /* set src addr (&sysGdt) */

#ifdef _WRS_CONFIG_SMP
        mov     FUNC(glbLoApicIdAdrs),%rdi
        movl    0x0(%rdi),%edi
        shrl    $_X86_64_CPU_INDEX_SHIFT,%edi
        movzb   FUNC(mpCpuIndexTable)(%rdi),%rdi
        subl    FUNC(vxBaseCpuPhysIndex),%edi

        _X86_64_PER_CPU_VALUE_GET_LONG(%rdi,%rax,pSysGdt)
#else
        mov     FUNC(pSysGdt), %rdi
#endif

        mov     %rdi,%rax
        mov     $GDT_ENTRIES,%rcx       /* number of GDT entries */
        mov     %rcx,%rdx
        shl     $1,%rcx                 /* set (nLongs of GDT) to copy */
        cld
        rep
        movsl                           /* copy GDT from src to dst */

        /* create a GDT descriptor on the stack */

        push    %rax                    /* push the (GDT base addr) */
        shl     $3,%rdx                 /* get (nBytes of GDT) */
        dec     %rdx                    /* get (nBytes of GDT) - 1 */
        shlq    $48,%rdx                /* move it to the upper 16 of %rex */
        push    %rdx                    /* push the nBytes of GDT - 1 as 64 bit */
        lea     6(%rsp),%rdi            /* get the addr of (size:addr): that */
                                        /* the start of the GDTPTR structure */
                                        /* push it as a parameter */

        call    FUNC(sysLoadGdt)        /* load the brand new GDT in RAM */

        mov     %rbp,%rsp
        popal                           /* re-initialize frame pointer */
        ret

/*******************************************************************************
*
* sysLoadGdt - load the global descriptor table.
*
* RETURNS: N/A
*
* NOMANUAL

* void sysLoadGdt (char *sysGdtr)
 
*/

        .balign 16,0x90
FUNC_LABEL(sysLoadGdt)
        lgdt    (%rdi)
        movw    $DATA_SEGMENT,%ax       /* a selector 0x10 is 3rd one */
	movw	%ax,%ds	
	movw	%ax,%es
	movw	%ax,%fs
	movw	%ax,%ss

#ifndef _WRS_CONFIG_SMP
    movw    %ax,%gs
#endif /* !_WRS_CONFIG_SMP */

	ret

/*******************************************************************************
*
* sysInByte - input one byte from I/O space
*
* SYNOPSIS
* \ss
* UCHAR sysInByte (address)
*     int address;	/@ I/O port address @/
* \se
*
* RETURNS: Byte data from the I/O port.
*
*/

	.balign 16,0x90
FUNC_LABEL(sysInByte)
        movl    %edi,%edx
	movl	$0,%eax
	inb	%dx,%al
	jmp	sysInByte0
sysInByte0:
	ret

/*******************************************************************************
*
* sysInWord - input one word from I/O space
*
* SYNOPSIS
* \ss
* USHORT sysInWord (address)
*     int address;	/@ I/O port address @/
* \se
*
* RETURNS: Word data from the I/O port.
*
*/

	.balign 16,0x90
FUNC_LABEL(sysInWord)
        movl    %edi,%edx
	movl	$0,%eax
	inw	%dx,%ax
	jmp	sysInWord0
sysInWord0:
	ret

/*******************************************************************************
*
* sysInLong - input one long-word from I/O space
*
* SYNOPSIS
* \ss
* UINT32 sysInLong (address)
*     int address;	/@ I/O port address @/
* \se
*
* RETURNS: Long-Word data from the I/O port.
*
*/

	.balign 16,0x90
FUNC_LABEL(sysInLong)
        movl    %edi,%edx
	movl	$0,%eax
	inl	%dx,%eax
	jmp	sysInLong0
sysInLong0:
	ret

/*******************************************************************************
*
* sysOutByte - output one byte to I/O space
*
* SYNOPSIS
* \ss
* void sysOutByte (address, data)
*     int address;	/@ I/O port address @/
*     char data;	/@ data written to the port @/
* \se
*
* RETURNS: N/A
*
*/

	.balign 16,0x90
FUNC_LABEL(sysOutByte)
        movl    %edi,%edx
        movl    %esi,%eax
	outb	%al,%dx
	jmp	sysOutByte0
sysOutByte0:
	ret

/*******************************************************************************
*
* sysOutWord - output one word to I/O space
*
* SYNOPSIS
* \ss
* void sysOutWord (address, data)
*     int address;	/@ I/O port address @/
*     short data;	/@ data written to the port @/
* \se
*
* RETURNS: N/A
*
*/

	.balign 16,0x90
FUNC_LABEL(sysOutWord)
        movl    %edi,%edx
        movl    %esi,%eax
	outw	%ax,%dx
	jmp	sysOutWord0
sysOutWord0:
	ret

/*******************************************************************************
*
* sysOutLong - output one long-word to I/O space
*
* SYNOPSIS
* \ss
* void sysOutLong (address, data)
*     int address;	/@ I/O port address @/
*     int data;		/@ data written to the port @/
* \se
*
* RETURNS: N/A
*
*/

	.balign 16,0x90
FUNC_LABEL(sysOutLong)
        movl    %edi,%edx
        movl    %esi,%eax
	outl	%eax,%dx
	jmp	sysOutLong0
sysOutLong0:
	ret

/*******************************************************************************
*
* sysInWordString - input word string from I/O space
*
* SYNOPSIS
* \ss
* void sysInWordString (port, address, count)
*     int port;		/@ I/O port address @/
*     short *address;	/@ address of data read from the port @/
*     int count;	/@ count @/
* \se
*
* RETURNS: N/A
*
*/

	.balign 16,0x90
FUNC_LABEL(sysInWordString)
        movl    %edx,%ecx       /* %ecx = count */
        movl    %edi,%edx       /* %edx = port */
        mov     %rsi,%rdi       /* %rdi = address */
	cld
	rep
	insw	%dx,(%rdi)      /* *address = port */
	mov 	%rdi,%rax       /* rax = address */
	ret

/*******************************************************************************
*
* sysInLongString - input long string from I/O space
*
* SYNOPSIS
* \ss
* void sysInLongString (port, address, count)
*     int port;		/@ I/O port address @/
*     int *address;	/@ address of data read from the port @/
*     int count;	/@ count @/
* \se
*
* RETURNS: N/A
*
*/

	.balign 16,0x90
FUNC_LABEL(sysInLongString)
        movl    %edx,%ecx       /* %ecx = count */
        movl    %edi,%edx       /* %edx = port */
        mov     %rsi,%rdi       /* %rdi = address */
	cld
	rep
	insl	%dx,(%rdi)      /* *address = port */
	mov	%rdi,%rax       /* rax = address */
	ret

/*******************************************************************************
*
* sysOutWordString - output word string to I/O space
*
* SYNOPSIS
* \ss
* void sysOutWordString (port, address, count)
*     int port;		/@ I/O port address @/
*     short *address;	/@ address of data written to the port @/
*     int count;	/@ count @/
* \se
*
* RETURNS: N/A
*
*/

	.balign 16,0x90
FUNC_LABEL(sysOutWordString)
        movl    %edx,%ecx       /* %ecx = count */
        movl    %edi,%edx       /* %edx = port */
	cld
	rep
	outsw	(%rsi),%dx      /* port = *address */
	mov	%rsi,%rax       /* rax = address */
	ret

/*******************************************************************************
*
* sysOutLongString - output long string to I/O space
*
* SYNOPSIS
* \ss
* void sysOutLongString (port, address, count)
*     int port;		/@ I/O port address @/
*     int *address;	/@ address of data written to the port @/
*     int count;	/@ count @/
* \se
*
* RETURNS: N/A
*
*/

	.balign 16,0x90
FUNC_LABEL(sysOutLongString)
        movl    %edx,%ecx       /* %ecx = count */
        movl    %edi,%edx       /* %edx = port */
	cld
	rep
	outsl	(%rsi),%dx      /* port = *address */
	mov	%rsi,%rax       /* rax = address */
	ret

/*******************************************************************************
*
* sysWait - wait until the input buffer become empty
*
* wait until the input buffer become empty
*
* SYNOPSIS
* \ss
* void sysWait (void)
* \se
*
* RETURNS: N/A
*
*/

	.balign 16,0x90
FUNC_LABEL(sysWait)
	xorl	%ecx,%ecx
sysWait0:
	movl	$0x64,%edx		/* Check if it is ready to write */
	inb	%dx,%al
	andb	$2,%al
	loopnz	sysWait0
	ret

/*******************************************************************************
*
* sysReboot - reboot system
*
* RETURNS: N/A
*
* NOMANUAL
*
* void sysReboot ()
*
*/

	.balign 16,0x90
FUNC_LABEL(sysReboot)

	mov     $0x2,%rdi
	call	FUNC(sysToMonitor)		/* initiate cold boot */
	ret

/*******************************************************************************
*
* sysCalTSC - Calibrate TSC (Time Stamp Counter)
* This routine uses PIT running at mode 3 for a specified milisecond duration
* to get the TSC frequency.
* Note:
* 1. duration must be greater than 0, and no greater than 20.
* 2. to avoid unnecessary error, make sure 1000 can be be exactly divided by input
* duration.
*
* RETURNS: N/A
*
* NOMANUAL
* UINT32 sysCalTSC ( int duration )
*/

    .balign 16,0x90
FUNC_LABEL(sysCalTSC)
    mov %rdi, %rdx	               /* get duration */
    push %rbx                          /* save rbx */
    pushf

    /*
     * pre-compute the end tick:
     * PIT clockrate is 1193182, and it is a count down clock. Fill it with
     * 65535, and wait to computed count to get specific delay. Considering
     * all the IO read/write in the loop will consume about 5us in most
     * targets, so compensate 5us to the end PIT count (about 12 PIT ticks).
     */

    movl $2386, %eax
    mull %edx
    movl $0xffff, %edx
    subl %eax, %edx
    addl $12, %edx                  /* 12 PIT tick compensation */
    push %rdx                      /* save the value in stack */

    /* select counter2, working at mode 3 */

    movb $0xb6, %al
    outb %al, $0x43

    /* set the initial count value to be 65535 */

    movb $0xff, %al
    outb %al, $0x42
    outb %al, $0x42

    /* enable counter 2 clock */

    inb  $0x61, %al
    orb  $1, %al
    outb %al, $0x61

    /* read the start count value of TSC, make it serialized */

    lfence
    rdtsc
    movl %eax, %ebx                 /* save start count in ebx */
    pop %rdx

sysCalTSC0:

    movl $0x080, %eax
    outb %al, $0x43
    inb  $0x42, %al
    movl %eax, %ecx                 /* save lower 8bit */
    inb  $0x42, %al
    shl  $8, %eax                   /* lshift 8bit */
    addl %eax, %ecx

    cmpl %edx, %ecx                 /* compare PIT end count */
    jge  sysCalTSC0

    /* read the end count of TSC */

    lfence
    rdtsc
    subl %ebx, %eax
    movl %eax, %ebx                 /* save the result */

    /* disable timer 2 clock */

    inb  $0x61,%al
    andb $0xfe,%al
    outb %al,$0x61

    movl %ebx, %eax
    popf
    pop  %rbx                          /* restore rbx */
    ret

/**************************************************************************/

        .data
        .balign 32,0x90
FUNC_LABEL(sysCsSuper)
        .quad   CODE_SEGMENT            /* CS for supervisor mode task */
FUNC_LABEL(sysCsExc)
        .long   0x00000018              /* CS for exception */
FUNC_LABEL(sysCsInt)
        .long   0x00000020              /* CS for interrupt */
FUNC_LABEL(sysSsSuper)
        .long   0x00000010              /* SS for supervisor mode task */

