/* sysNvRam.c - x86 nvram api for block devices */

/*
 * Copyright (c) 2014 Wind River Systems, Inc.
 *
 * The right to copy, distribute, modify or otherwise make use
 * of this software may be licensed only pursuant to the terms
 * of an applicable Wind River license agreement.
 */

/*
modification history
--------------------
17dec14,j_l  V7PRO-1506 - Fix static analysis errors
04nov14,rbc  US44720 - Fix static analysis errors
18mar14,j_l  US32795 - restore sysNvRamGet/Set functionality
04feb14,scm  US24478 - initial mods for GEN2 test image...
31jan14,j_l  initial creation based on itl_64 version 01aug13
*/

/*
DESCRIPTION:

This file implements the nvram api using standard io system calls
for non-volatile media such as flash ram, floppy disks and hard disk
drives.  It is primarily intended for saving and modifying boot parameters
but could in theory be used for general purpose information.  A file of
NV_RAM_SIZE size is used for storing the "nvram" data.
*/

/* includes */

#include <vxWorks.h>
#include <logLib.h>
#include <string.h>
#ifdef PRJ_BUILD
#include "prjParams.h"
#endif /* PRJ_BUILD */

/* defines */

/*
 * Allow for multiple attempts to load a file.
 * May have to wait for Filesystem to mount.
 */

#define NUMBER_OF_ATTEMPTS     2
#define DELAY_BETWEEN_ATTEMPTS (5*sysClkRateGet ())

#define NVRAMFILE  "/nvram.txt"
#define NVRAMPATH   BOOTROM_DIR "" NVRAMFILE

/******************************************************************************
*
* sysNvRamGet - get the contents of non-volatile RAM
*
* This routine copies the contents of non-volatile memory into a specified
* string.  The string is terminated with an EOS.
*
* RETURNS: OK, or ERROR if access is outside the non-volatile RAM range.
*
* SEE ALSO: sysNvRamSet()
*/

STATUS sysNvRamGet
    (
    char *string,    /* where to copy non-volatile RAM    */
    int strLen,      /* maximum number of bytes to copy   */
    int offset       /* byte offset into non-volatile RAM */
    )
    {
    int fd=ERROR, bytes;
    int unsuccessfulAttempts = 0;

    if ((offset < 0) || (strLen < 0) || ((offset + strLen) > NV_RAM_SIZE))
        return (ERROR);

    while ((fd == ERROR) && (unsuccessfulAttempts < NUMBER_OF_ATTEMPTS))
        {
        if ((fd = open (NVRAMPATH, O_RDONLY, 0)) == ERROR)
            {
            if (++unsuccessfulAttempts < NUMBER_OF_ATTEMPTS)
                {

                /* return value of logMsg() not used */

                (void ) logMsg ("sysNvRamGet: '%s' open failed.  "
                                "Attempt %x of %x.\n",
                                (_Vx_usr_arg_t)NVRAMPATH,
                                (_Vx_usr_arg_t)unsuccessfulAttempts,
                                (_Vx_usr_arg_t)NUMBER_OF_ATTEMPTS,
                                0, 0, 0);

                /* return value of taskDelay() not used */

                (void) taskDelay (DELAY_BETWEEN_ATTEMPTS);

                }
            else
                {

                /* return value of logMsg() not used */

                (void) logMsg ("sysNvRamGet: '%s' open failed\n",
                               (_Vx_usr_arg_t)NVRAMPATH, 0, 0, 0, 0, 0);
                return (ERROR);

                }
            }
        }

    if (lseek (fd, offset, SEEK_SET) != offset)
        {

        /* return value of close() not used */

        (void) close (fd);
        return (ERROR);

        }

    bytes = read (fd, string, strLen);
    if (bytes == ERROR)
        {

        /* return value of logMsg() not used */

        (void) logMsg ("sysNvRamGet: '%s' read failed\n",
                       (_Vx_usr_arg_t)NVRAMPATH, 0, 0, 0, 0, 0);

        /* return value of close() not used */

        (void) close (fd);
        return (ERROR);

        }
    else
        {
        string[bytes] = EOS;

        /* return value of close() not used */

        (void) close (fd);
        return (OK);
        }
    }

/*******************************************************************************
*
* sysNvRamSet - write to non-volatile RAM
*
* This routine copies a specified string into non-volatile RAM.
*
* RETURNS: OK, or ERROR if access is outside the non-volatile RAM range.
*
* SEE ALSO: sysNvRamGet()
*/

STATUS sysNvRamSet
    (
    char *string,     /* string to be copied into non-volatile RAM */
    int strLen,       /* maximum number of bytes to copy           */
    int offset        /* byte offset into non-volatile RAM         */
    )
    {
    int fd;

    if ((offset < 0) || (strLen < 0) || ((offset + strLen) > NV_RAM_SIZE))
        return (ERROR);

    fd = open (NVRAMPATH, O_RDWR | O_CREAT, DEFFILEMODE);
    if (fd == ERROR)
        {

        /* return value of logMsg() not used */

        (void) logMsg ("sysNvRamSet: '%s' open failed\n",
                       (_Vx_usr_arg_t)NVRAMPATH, 0, 0, 0, 0, 0);
        return (ERROR);

        }

    if (lseek (fd, offset, SEEK_SET) != offset)
        {

        /* return value of close() not used */

        (void) close (fd);
        return (ERROR);
        }

    if (write (fd, string, strLen) != strLen)
        {

        /* return value of logMsg() not used */

        (void) logMsg ("sysNvRamSet: '%s' write failed\n",
                       (_Vx_usr_arg_t)NVRAMPATH, 0, 0, 0, 0, 0);

        /* return value of close() not used */

        (void) close (fd);
        return (ERROR);

        }
    else
        {

        /* return value of close() not used */

        (void) close (fd);
        return (OK);

        }
    }


