/* vxbAccesDio.h - ACCES I/O Products DIO driver header file*/

/* Copyright (c) 2018, ACCES I/O Products, Inc.
*
* The right to copy, distribute, modify or otherwise make use of this 
* software may be licensed only pursuant to the terms  of an applicable 
* ACCES I/O Products, Inc., license agreement.
*/

/*
modification history
--------------------
21jun18,acc  First release.
*/

/*
DESCRIPTION
===========
This header file defines and documents the user interface for developers using 
various ACCES I/O Product, Inc., digital I/O cards, on a PCI bus, including the 
mPCIe-DIO-24S. It also introduces the features of the hardware and provides
references for the hardware and PCI. (NOTE: in this context, "user" refers to
a VxWorks task calling an API.)

This version of the driver supports:

- ACCES products that emulate aspects of an Intel 8255 extended to provide an 
  interrupt when a pin Change-of-State (CoS) occurs. Hardware features include:
  - 24 I/O pins in four ports: two 8-bit ports, Port A and Port B, and two four-
    bit ports, Port CHi and Port CLo.
  - An 8-bit "Control Word register" (8255 datasheet terminology, though it is 
    one byte) used to set direction for each of the four ports. The Control
    Word is *not* to be accessed directly by the user.  
  - Programmable in I/O Mode 0 (Basic Input/Output) only (see 8255 reference
    below).

- Read/write access to the four ports only as a single UINT32--all of Ports A,
  B, CHi, and CLo are read from/written to at once using a UINT32, in little-
  endian order, as shown below. Within each byte, bit 0 is the least significant 
  bit of that byte. (The "Control Word" is shown but generally should not be
  accessed by the user.)

                MSB                                 LSB
               -----------------------------------------
           Bit | 31 - 24 | 23 - 16 | 15 -  8 |  7 -  0 |
               -----------------------------------------
      Register | Control | CHi|CLo |  Port B |  Port A |
               -----------------------------------------

  All outputs are latched and may be read repeatedly without changing state.
  
  NOTE: No type is defined for this UINT32.

- A single ACCES DIO card in the system, with device name /accesDio/0. To access
  the device, a user task must call accesDioOpen() to reserve it, giving that
  task exclusive access until a call to accesDioClose(). Thus, no special
  considerations are required to avoid conflict among multiple tasks or CPUs. 
  
References
----------
Further details for DIO devices may be found in four sources:

- This file vxbAccesDio.h: #defines for port masks, port access macros, and
  comments for each user-callable functions prototype.

- ACCES I/O Products, Inc. "24 Digital I/O with CoS IRQs for Mini PCI Express 
  Hardware Manual." Version: mPCIe-DIO-24S Family Manual, Rev 82a.
  http://www.accesio.com/MANUALS/mPCIe-DIO-24S.pdf
  
- Intersil Datasheet, "82C55A CMOS Programmable Peripheral Interface," FN269,
  Rev 11.00, Renesys, Dec 8, 2015. (Best formatting, searchable.)
  https://www.intersil.com/content/dam/Intersil/documents/82c5/82c55a.pdf.
  
- Wikipedia. "Intel 8255." Last edited 29 April 2018.
  https://en.wikipedia.org/wiki/Intel_8255

For information on PCI, the following Wikipedia article and related articles
are a useful introduction:

 - Wikipedia. "PCI configuration space." Last edited 6 March 2018.
   https://en.wikipedia.org/wiki/PCI_configuration_space.
*/

#ifndef __INCaccesDioDrvh
#define __INCaccesDioDrvh

#ifdef __cplusplus
extern "C" {
#endif

/* Includes */

#include <vxWorks.h>
                        
/*******************************************************************************
* Types and Defines
* ================= 
*/

/*
 * Port direction values passed to accesDioConfigure(): a port must be set to 
 * output direction before writing to it, and set to input direction before 
 * reading the state of the pins.
 * 
 * NOTE: The concepts of "port direction" and of "read/write" to a port using
 * the functions provided here must be considered together:
 * - If a port is set to input direction, a write to it is ignored.
 * - If a port is set to output direction, a read from it will return the value 
 *   last written to it. 
 */

typedef enum port_dir
    {
    PORT_DIR_INPUT,     /* set port to get data from external devices */
    PORT_DIR_OUTPUT     /* set port to send data to external devices */
    } PORT_DIR;

/* Masks for port values and Control Word within a UINT32. */

#define PORT_A_MASK     0x000000ff
#define PORT_B_MASK     0x0000ff00
#define PORT_C_MASK     0x00ff0000  /* To treat Ports CHi and CLo together. */
#define PORT_CLO_MASK   0x000f0000
#define PORT_CHI_MASK   0x00f00000
#define CONTROL_MASK    0xff000000

/*
 * Macros to extract a port value from a UINT32 to a one-byte value and to set
 * a one-byte value to the correct bits within a UIN32 for a port. The macros 
 * "extract" and "insert" within a UINT32, but do NOT read or write the device.
 *  
 * For Ports CHi and CLo:
 * 
 * PORT_CLO_EXTRACT(u32)     Evaluate to a byte with the CLo bits from u32; 
 *                           the *upper* four bits will be 0.
 * PORT_CLO_INSERT(u32, u8)  Set CLo bits within u32 from the lower four bits of
 *                           byte u8 (the upper four bits are don't care).
 * PORT_CHI_EXTRACT(u32)     Evaluate to a byte with the CHi bits from u32; the 
 *                           *lower* four bits will be 0 (the value is not 
 *                           shifted).
 * PORT_CHI_INSERT(u32, u8)  Set CHi bits within u32 from the *upper* four bits
 *                           of byte u8 (the lower four bits are don't care).
 * 
 * The SET macros:
 * - Leave all bits in the UINT32 value unchanged except for the bits of the 
 *   given port. 
 * - Evaluate to a UINT32 value, which may then be used, for example, as an
 *   argument to accesDioWritePorts().  
 */

#define PORT_A_EXTRACT(u32)      (UINT8)   (u32 & PORT_A_MASK) 
#define PORT_B_EXTRACT(u32)      (UINT8)  ((u32 & PORT_B_MASK) >> 8)
#define PORT_C_EXTRACT(u32)      (UINT8)  ((u32 & PORT_C_MASK) >> 16)
#define PORT_CLO_EXTRACT(u32)    (UINT8)  ((u32 & PORT_CLO_MASK) >> 16)
#define PORT_CHI_EXTRACT(u32)    (UINT8)  ((u32 & PORT_CHI_MASK) >> 16)

#define PORT_A_INSERT(u32, u8)   (UINT32) ((u32 & ~PORT_A_MASK) | u8)
#define PORT_B_INSERT(u32, u8)   (UINT32) ((u32 & ~PORT_B_MASK) | (u8 << 8)
#define PORT_CLO_INSERT(u32, u8) (UINT32)(((u32 & ~PORT_CLO_MASK) | ((u8 << 16) & PORTA_CLO_MASK))
#define PORT_CHI_INSERT(u32, u8) (UINT32)(((u32 & ~PORT_CHI_MASK) | ((u8 << 16) & PORTA_CHI_MASK))

/*******************************************************************************
* User-Callable Function Prototypes
* =================================
* This driver provides the following user-callable functions (presented 
* alphabetically in standard VxWorks format). There are no higher-level 
* functions such as open() and ioctl().
*/

/*******************************************************************************
*
* accesDioClose - close the DIO card and make it available to another user
*
* This function closes the card and makes it available to another task. It is 
* not an error to close the card if it is not open. It is an error for a task
* to attempt to close an open the card if it is not the owner (that is, if it is 
* not the task that opened it). Closing the card disables interrupts if enabled,
* and always clears any pending interrupts and resets the ISR callback function 
* pointer to NULL.
* 
* RETURNS: OK or ERROR per errno.
* 
* ERRNO
*   ENXIO       Driver not installed.
*   EACCES      Card is open for another task.
*/

STATUS accesDioClose (void);

/*******************************************************************************
*
* accesDioConfigure - set the direction of each port to input or output
*
* Set the direction of each port either to get input from the pins or put output  
* to the pins based on its corresponding argument. The arguments must be from 
* the PORT_DIR enumeration defined in the vxbAccesDio.h header file 
* (PORT_DIR_INPUT or PORT_DIR_OUTPUT). See the note regarding "port direction" 
* and "read/write" to a port with the definition of PORT_DIR.
* 
* NOTE: A port must be set to input direction to generate CoS interrupts. See
* accesDioIntEnable() for more.
* 
* RETURNS: OK or ERROR per errno.
* 
* ERRNO
*   ENXIO       Driver not installed.
*   ENODEV      Card is not open.
*   EACCES      Card is open for another task.
*/
 
STATUS accesDioConfigure
    (
    PORT_DIR    portADir,   /* Port A direction */
    PORT_DIR    portBDir,   /* Port B direction */
    PORT_DIR    portCLoDir,  /* Port DLo direction */
    PORT_DIR    portCHiDir /* Port CHi direction */
    );

/*******************************************************************************
*
* accesDioIntCallbackInstall - install the user interrupt service routine (ISR)
*
* Record the callback interrupt service routine (ISR) function pointer argument.
* The prototype for that function must be:
* 
*   STATUS <ISR function name> (void)
* 
* The device must be open by calling accesDioOpen() before calling this routine.
* 
* The ISR will be called if:
* - This routine has been called to establish the ISR callback function pointer.
* - Interrupts are enabled by calling accesDioIntEnable().
* - There is a change of state (CoS) on a pin on a port configured for input
*   while interrupts are enabled.
* 
* CoS will be ignored if interrupts are not then enabled, and are ignored for 
* pins on a port set for output direction.
* 
* This routine may be called at any time that the card is open to change the 
* callback function pointer. It may be called with NULL function pointer to
* remove the callback function.
*
* RETURNS: OK or ERROR per errno.
* 
* ERRNO:
*   ENXIO       Driver not installed.
*   ENODEV      Card is not open.
*   EACCES      Card is open for another task.
*/

STATUS accesDioIntCallbackInstall
    (
    STATUS  (*callback_funcptr)()   /* callback function pointer */
    );

/******************************************************************************
*
* accesDioIntClear - clear a pending interrupt, if any
*
* If an interrupt is pending, this function clears it; otherwise there is no
* effect. NOTE: It is unusual for a user task to call this function because
* interrupts are cleared as part of the interrupt service routine.
* 
* RETURNS: OK or ERROR per errno.
* 
* ERRNO
*   ENXIO       Driver not installed.
*   ENODEV      Card is not open.
*   EACCES      Card is open for another task.
*/

STATUS accesDioIntClear (void);

/******************************************************************************
*
** accesDioIntDisable - disable interrupts from the card
*
* This routine disables interrupts from the card. Any pending interrupts are
* cleared. Calling this routine when interrupts are disabled is not an error.
* 
* RETURNS: OK or ERROR per errno.
* 
* ERRNO
*   ENXIO       Driver not installed.
*   ENODEV      Card is not open.
*   EACCES      Card is open for another task.
*/

STATUS accesDioIntDisable (void);

/******************************************************************************
*
* accesDioIntEnable - enable interrupts from the card
*
* This routine enables Change-of-State (CoS) interrupts from the card. A CoS
* interrupt occurs whenever any input pin changes state due to an external 
* event. Calling this function when interrupts are enabled is not an error. 
* Interrupts are not generated for ports set for output.
* 
* RETURNS: OK or ERROR per errno.
* 
* ERRNO
*   ENXIO       Driver not installed.
*   ENODEV      Card is not open.
*   EACCES      Card is open for another task.
*/

STATUS accesDioIntEnable (void);

/*******************************************************************************
*
* accesDioIntPending - test whether an interrupt is pending from the card
*
* This routine is passed a pointer to a byte. It sets the byte to 1 if an
* interrupt was generated and not yet cleared (pending) and 0 if not. Reading 
* the interrupt status does not clear it; call accesDioIntClear() to clear the 
* interrupt.
* 
* RETURNS: OK or ERROR per errno.
* 
* ERRNO
*   ENXIO       Driver not installed.
*   ENODEV      Card is not open.
*   EACCES      Card is open for another task.
*   EINVAL      Null argument pointer.
*
* NOTE: Because interrupts are cleared as part of the interrupt service routine
* in this driver, accesDioIntPending() could set *pIntStatus to TRUE only if one
* or more of the conditions required for the ISR to be invoked on an interrupt 
* were not met.
*/

STATUS accesDioIntPending
    (
    BOOL *  pIntStatus   /* pointer to byte to store the interrupt status: 
                            1 if an interrupt is pending, 0 if not */
    );

/*******************************************************************************
*
* accesDioOpen - open the DIO card and assign it to the user
*
* This version of this driver supports a single DIO card in the system. This
* function opens that card and assigns it to the calling user task. A DIO card 
* must be opened before it can be used. It stays open for the exclusive use of 
* the calling user until it is freed by calling acessDioClose().
* 
* It is an error to call this function if the card is already opened for either
* the current task or another task.
*                         
* RETURNS: OK or ERROR per errno.
* 
* ERRNO
*   ENXIO       Driver not installed.
*   ENODEV      Card is already open for this task.
*   EACCES      Card is open for another task.
*/

STATUS accesDioOpen (void);

/*******************************************************************************
* accesDioReadControl - read the Control Word (byte) from the device
*
* This function reads the Control Word from the device and stores it in the 
* UINT8 location passed to it. While available, it would ordinarily not be 
* called by a user.
*
* RETURNS: OK or ERROR per errno.
* 
* ERRNO
*   ENXIO       Driver not installed.
*   ENODEV      Card is not open.
*   EACCES      Card is open for another task.
*   EINVAL      Null argument pointer.
*/

STATUS accesDioReadControl
    (
    UINT8 * pControl    /* pointer to location to store the Control Word */
    );

/*******************************************************************************
*
* accesDioReadPorts - read all of Ports A, B, CHi, and CLo
*
* This function is passed a pointer to a UINT32; it reads Ports A, B, CHi, and 
* CLo and stores the values in the three low-order bytes of the UINT32 (see 
* aioDioWritePorts() for the position of each port in the UINT32). The MSB of  
* the UINT32 is set to 0x00.
* 
* RETURNS: OK or ERROR per errno.
* 
* ERRNO
*   ENXIO       Driver not installed.
*   ENODEV      Card is not open.
*   EACCES      Card is open for another task.
*   EINVAL      Null argument pointer.
*/

STATUS accesDioReadPorts
    (
    UINT32 *    pValue      /* pointer to location to store the three port
                               values read from the device; the MSB is set to
                               0x00 */
    );

/******************************************************************************
*
* accesDioShow - show driver and device details for the DIO card
*
* Show the current driver and device details for the single DIO card supported
* in this version of this driver. If the driver is not initialized or the card
* is not attached, the function reports that but does not return ERROR.
* 
* For output, uses _func_kwrite (likely either kwriteVga() or kwriteSerial())
* if it is setup, or printf() if not.
* 
* RETURNS: N/A
* 
* ERRNO: N/A
*/

void accesDioShow (void);

/*******************************************************************************
*
* accesDioWritePorts - write to all of Ports A, B, CHi, and CLo
*
* This function is passed a single UINT32; it writes the low-order 24 bits to
* Ports A, B, CHi, and CLo. Port A is 0x000000FF, Port B is 0x0000FF00, Port CLo 
* is 0x000F0000, and Port CHi is 0x00F00000. The MSB of the argument is ignored. 
* Masks PORT_A_MASK .. PORT_C_MASK (CHi and CLo together) are defined in header 
* file vxbAccesDio.h.
*
* RETURNS: OK or ERROR per errno.
* 
* ERRNO
*   ENXIO       Driver not installed.
*   ENODEV      Card is not open.
*   EACCES      Card is open for another task.
*/
 
STATUS accesDioWritePorts
    (
    UINT32  ports           /* value to be written */
    );



#ifdef __cplusplus
}
#endif

#endif  /* __INCaccesDioDrvh */
